from django import forms as djangoforms
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.forms import fields as djangofields
from django_sorcery import fields as sorceryfields, forms

from .base import TestCase
from .testapp.models import (
    ClassicModel,
    ModelFullCleanFail,
    ModelOne,
    ModelTwo,
    Option,
    Owner,
    Part,
    Vehicle,
    VehicleType,
    db,
)


class TestFieldsForModel(TestCase):
    def test_default_case(self):
        fields = forms.fields_for_model(Vehicle, db)

        checks = [
            ("name", djangofields.CharField),
            ("type", sorceryfields.EnumField),
            ("created_at", djangofields.DateTimeField),
            ("paint", djangofields.TypedChoiceField),
            ("is_used", djangofields.BooleanField),
            ("msrp", djangofields.DecimalField),
            ("owner", sorceryfields.ModelChoiceField),
            ("parts", sorceryfields.ModelMultipleChoiceField),
            ("options", sorceryfields.ModelMultipleChoiceField),
        ]
        self.assertEqual(len(fields), len(checks))
        for name, fieldtype in checks:
            self.assertIsInstance(fields[name], fieldtype)

    def test_fields(self):
        fields = forms.fields_for_model(Vehicle, db, fields=("name",))
        checks = [("name", djangofields.CharField)]
        self.assertEqual(len(fields), len(checks))
        for name, fieldtype in checks:
            self.assertIsInstance(fields[name], fieldtype)

    def test_exclude(self):
        fields = forms.fields_for_model(Vehicle, db, exclude=("name",))
        checks = [
            ("type", sorceryfields.EnumField),
            ("created_at", djangofields.DateTimeField),
            ("paint", djangofields.TypedChoiceField),
            ("is_used", djangofields.BooleanField),
            ("msrp", djangofields.DecimalField),
            ("owner", sorceryfields.ModelChoiceField),
            ("parts", sorceryfields.ModelMultipleChoiceField),
            ("options", sorceryfields.ModelMultipleChoiceField),
        ]
        self.assertEqual(len(fields), len(checks))
        for name, fieldtype in checks:
            self.assertIsInstance(fields[name], fieldtype)

    def test_widgets(self):
        fields = forms.fields_for_model(Vehicle, db, fields=("name",), widgets={"name": djangoforms.Textarea})

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.CharField)
        self.assertIsInstance(fields["name"].widget, djangoforms.Textarea)

    def test_localized_fields(self):
        fields = forms.fields_for_model(Vehicle, db, fields=("name",), localized_fields=("name",))

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.CharField)
        self.assertTrue(fields["name"].localize)

        fields = forms.fields_for_model(Vehicle, db, fields=("name",), localized_fields=djangoforms.ALL_FIELDS)

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.CharField)
        self.assertTrue(fields["name"].localize)

    def test_labels(self):
        fields = forms.fields_for_model(Vehicle, db, fields=("name",), labels={"name": "dummy"})

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.CharField)
        self.assertEqual(fields["name"].label, "dummy")

    def test_help_texts(self):
        fields = forms.fields_for_model(Vehicle, db, fields=("name",), help_texts={"name": "dummy"})

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.CharField)
        self.assertEqual(fields["name"].help_text, "dummy")

    def test_error_messages(self):
        fields = forms.fields_for_model(Vehicle, db, fields=("name",), error_messages={"name": {"required": "dummy"}})

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.CharField)
        self.assertEqual(fields["name"].error_messages, {"required": "dummy"})

    def test_field_classes(self):
        fields = forms.fields_for_model(
            Vehicle, db, fields=("name",), help_texts={"name": None}, field_classes={"name": djangofields.FileField}
        )

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.FileField)

    def test_formfield_callback(self):
        def callback(*args, **kwargs):
            return djangofields.FileField()

        fields = forms.fields_for_model(Vehicle, db, fields=("name",), formfield_callback=callback)

        self.assertEqual(len(fields), 1)
        self.assertIsInstance(fields["name"], djangofields.FileField)

        with self.assertRaises(TypeError):
            fields = forms.fields_for_model(Vehicle, db, fields=("name",), formfield_callback=[])


class TestApplyLimitChoicesTo(TestCase):
    def test_apply_limit(self):
        db.add_all([Owner(first_name="one"), Owner(first_name="one_more"), Owner(first_name="two")])
        field = sorceryfields.ModelChoiceField(Owner, db, limit_choices_to=[Owner.first_name.startswith("one")])
        forms.apply_limit_choices_to_form_field(field)
        self.assertEqual(field.queryset.count(), 2)


class TestModelToDict(TestCase):
    def test_model_to_dict(self):
        vehicle = Vehicle(
            id=1,
            name="vehicle",
            owner=Owner(id=2, first_name="first_name", last_name="last_name"),
            is_used=True,
            paint="red",
            type=VehicleType.car,
            options=[Option(id=3, name="option 1"), Option(id=4, name="option 2")],
            parts=[Part(id=5, name="part 1"), Part(id=6, name="part 2")],
        )

        self.assertEqual(
            {
                "created_at": None,
                "is_used": True,
                "msrp": None,
                "name": "vehicle",
                "options": [3, 4],
                "owner": 2,
                "paint": "red",
                "parts": [5, 6],
                "type": VehicleType.car,
            },
            forms.model_to_dict(vehicle),
        )

    def test_model_to_dict_exclude(self):
        vehicle = Vehicle(
            id=1,
            name="vehicle",
            owner=Owner(id=2, first_name="first_name", last_name="last_name"),
            is_used=True,
            paint="red",
            type=VehicleType.car,
            options=[Option(id=3, name="option 1"), Option(id=4, name="option 2")],
            parts=[Part(id=5, name="part 1"), Part(id=6, name="part 2")],
        )

        self.assertEqual(
            {
                "created_at": None,
                "is_used": True,
                "msrp": None,
                "name": "vehicle",
                "options": [3, 4],
                "paint": "red",
                "parts": [5, 6],
            },
            forms.model_to_dict(vehicle, exclude=["type", "owner"]),
        )

    def test_model_to_dict_fields(self):
        vehicle = Vehicle(
            name="vehicle",
            owner=Owner(first_name="first_name", last_name="last_name"),
            is_used=True,
            paint="red",
            type=VehicleType.car,
            options=[Option(name="option 1"), Option(name="option 2")],
            parts=[Part(name="part 1"), Part(name="part 2")],
        )

        self.assertEqual(
            {"is_used": True, "name": "vehicle", "paint": "red"},
            forms.model_to_dict(vehicle, fields=["name", "is_used", "paint"]),
        )

    def test_model_to_dict_private_relation(self):
        obj = ModelTwo(pk=2, name="two", _model_one=ModelOne(pk=1, name="one"))

        self.assertEqual({"name": "two"}, forms.model_to_dict(obj))


class TestModelForm(TestCase):
    def setUp(self):
        super().setUp()
        self.owner = Owner(first_name="Test", last_name="Owner")
        db.add(self.owner)

        self.options = [
            Option(id=1, name="Option 1"),
            Option(id=2, name="Option 2"),
            Option(id=3, name="Option 3"),
            Option(id=4, name="Option 4"),
        ]
        db.add_all(self.options)
        db.flush()

    def test_modelform_factory_fields(self):
        form_class = forms.modelform_factory(Vehicle, fields=forms.ALL_FIELDS, session=db)
        form = form_class()
        self.assertListEqual(
            sorted(form.fields.keys()),
            ["created_at", "is_used", "msrp", "name", "options", "owner", "paint", "parts", "type"],
        )

    def test_modelform_factory_instance_validate(self):
        vehicle = Vehicle(owner=self.owner)
        form_class = forms.modelform_factory(Vehicle, fields=forms.ALL_FIELDS, session=db)
        form = form_class(instance=vehicle, data={"name": "testing"})
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {"type": ["This field is required."], "is_used": ["This field is required."]})

    def test_modelform_factory_instance_save(self):
        form_class = forms.modelform_factory(Vehicle, fields=forms.ALL_FIELDS, session=db)
        data = {"name": "testing", "type": "car", "owner": self.owner.id, "is_used": True}
        form = form_class(data=data)
        self.assertTrue(form.is_valid(), form.errors)
        form.save()

    def test_modelform_factory_modelchoicefield_choices(self):
        form_class = forms.modelform_factory(Vehicle, fields=forms.ALL_FIELDS, session=db)
        data = {"name": "testing", "type": "car", "owner": 1}
        form = form_class(data=data)

        owner_choices = form.fields["owner"].choices
        option_choices = form.fields["options"].choices
        self.assertEqual(len(owner_choices), 2)
        self.assertEqual(len(option_choices), 4)

    def test_modelform_factory_new_render(self):
        form_class = forms.modelform_factory(Vehicle, fields=forms.ALL_FIELDS, session=db)
        form = form_class(data={})

        self.assertTrue(form.is_bound)
        self.assertEqual(form.errors, {"type": ["This field is required."], "is_used": ["This field is required."]})
        self.assertEqual(
            form.initial,
            {"paint": None, "created_at": None, "type": None, "msrp": None, "name": None, "is_used": False},
        )
        self.assertEqual(
            form.cleaned_data,
            {"paint": "", "created_at": None, "options": [], "parts": [], "msrp": None, "name": "", "owner": None},
        )

        form.order_fields(sorted(form.fields.keys()))

        soup = self.get_soup(form.as_p())
        expected_soup = self.get_soup(
            "".join(
                [
                    "<p>",
                    '  <label for="id_created_at">Created at:</label>',
                    '  <input type="text" name="created_at" id="id_created_at" />',
                    "</p>",
                    '<ul class="errorlist"><li>This field is required.</li></ul>',
                    "<p>",
                    '  <label for="id_is_used">Is used:</label>',
                    '  <input id="id_is_used" name="is_used" required="" type="checkbox"/>',
                    "</p>",
                    "<p>",
                    '  <label for="id_msrp">Msrp:</label>',
                    '  <input id="id_msrp" name="msrp" step="0.01" type="number"/>',
                    "</p>",
                    "<p>",
                    '  <label for="id_name">Name:</label>',
                    '  <input id="id_name" maxlength="50" name="name" type="text" />',
                    '  <span class="helptext">' "    The name of the vehicle",
                    "  </span>",
                    "</p>",
                    "<p>",
                    '  <label for="id_options">Options:</label>',
                    '  <select id="id_options" multiple="multiple" name="options">',
                    "    <option value=\"1\">Option(id=1, name='Option 1')</option>",
                    "    <option value=\"2\">Option(id=2, name='Option 2')</option>",
                    "    <option value=\"3\">Option(id=3, name='Option 3')</option>",
                    "    <option value=\"4\">Option(id=4, name='Option 4')</option>",
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_owner">Owner:</label>',
                    '  <select id="id_owner" name="owner">',
                    "    <option selected value>---------</option>",
                    f'    <option value="{self.owner.id}">{self.owner}</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_paint">Paint:</label>',
                    '  <select id="id_paint" name="paint">',
                    "    <option selected value></option>",
                    '    <option value="red">red</option>',
                    '    <option value="green">green</option>',
                    '    <option value="blue">blue</option>',
                    '    <option value="silver">silver</option>',
                    '    <option value="pink">pink</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_parts">Parts:</label>',
                    '  <select id="id_parts" multiple="multiple" name="parts">',
                    "  </select>",
                    "</p>",
                    '<ul class="errorlist">',
                    "  <li>",
                    "    This field is required.",
                    "  </li>",
                    "</ul>",
                    "<p>",
                    '  <label for="id_type">Type:</label>',
                    '  <select id="id_type" name="type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                ]
            )
        )
        self.assertEqual(soup.prettify(), expected_soup.prettify())

    def test_modelform_factory_instance_render(self):
        form_class = forms.modelform_factory(Vehicle, fields=forms.ALL_FIELDS, session=db)
        vehicle = Vehicle(owner=self.owner, type=VehicleType.car)
        form = form_class(instance=vehicle, data={})

        self.assertTrue(form.is_bound)
        self.assertEqual(form.errors, {"type": ["This field is required."], "is_used": ["This field is required."]})
        self.assertEqual(
            form.initial,
            {
                "created_at": None,
                "is_used": False,
                "msrp": None,
                "name": None,
                "owner": self.owner.id,
                "paint": None,
                "type": VehicleType.car,
            },
        )
        self.assertEqual(
            form.cleaned_data,
            {"paint": "", "msrp": None, "created_at": None, "options": [], "parts": [], "name": "", "owner": None},
        )

        form.order_fields(sorted(form.fields.keys()))

        soup = self.get_soup(form.as_p())
        expected_soup = self.get_soup(
            "".join(
                [
                    "<p>",
                    '  <label for="id_created_at">Created at:</label>',
                    '  <input type="text" name="created_at" id="id_created_at" />',
                    "</p>",
                    '<ul class="errorlist"><li>This field is required.</li></ul>',
                    "<p>",
                    '  <label for="id_is_used">Is used:</label>',
                    '  <input id="id_is_used" name="is_used" required type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_msrp">Msrp:</label>',
                    '  <input id="id_msrp" name="msrp" step="0.01" type="number" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_name">Name:</label>',
                    '  <input id="id_name" maxlength="50" name="name" type="text" />',
                    '  <span class="helptext">' "    The name of the vehicle",
                    "  </span>",
                    "</p>",
                    "<p>",
                    '  <label for="id_options">Options:</label>',
                    '  <select id="id_options" multiple="multiple" name="options">',
                    "    <option value=\"1\">Option(id=1, name='Option 1')</option>",
                    "    <option value=\"2\">Option(id=2, name='Option 2')</option>",
                    "    <option value=\"3\">Option(id=3, name='Option 3')</option>",
                    "    <option value=\"4\">Option(id=4, name='Option 4')</option>",
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_owner">Owner:</label>',
                    '  <select id="id_owner" name="owner">',
                    "    <option selected value>---------</option>",
                    f'    <option value="{self.owner.id}">{self.owner}</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_paint">Paint:</label>',
                    '  <select id="id_paint" name="paint">',
                    "    <option selected value></option>",
                    '    <option value="red">red</option>',
                    '    <option value="green">green</option>',
                    '    <option value="blue">blue</option>',
                    '    <option value="silver">silver</option>',
                    '    <option value="pink">pink</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_parts">Parts:</label>',
                    '  <select id="id_parts" multiple="multiple" name="parts">',
                    "  </select>",
                    "</p>",
                    '<ul class="errorlist">',
                    "  <li>",
                    "    This field is required.",
                    "  </li>",
                    "</ul>",
                    "<p>",
                    '  <label for="id_type">Type:</label>',
                    '  <select id="id_type" name="type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                ]
            )
        )

        self.assertHTMLEqual(soup.prettify(), expected_soup.prettify())

    def test_form_field_callback_in_base_meta(self):
        self.callback_called = False

        def callback(*args, **kwargs):
            self.callback_called = True

        class OwnerBaseForm(forms.ModelForm):
            class Meta:
                model = Owner
                session = db
                fields = forms.ALL_FIELDS
                formfield_callback = staticmethod(callback)

        class OwnerForm(OwnerBaseForm):
            pass

        self.assertTrue(self.callback_called)

    def test_fields_bad_value(self):
        with self.assertRaises(TypeError) as ctx:
            forms.modelform_factory(Owner, forms.ModelForm, fields="abc1234")

        self.assertEqual(
            ctx.exception.args, ("OwnerForm.Meta.fields cannot be a string. Did you mean to type: ('abc1234',)?",)
        )

    def test_empty_fields_and_exclude(self):
        with self.assertRaises(ImproperlyConfigured) as ctx:

            class OwnerForm(forms.ModelForm):
                class Meta:
                    model = Owner

        self.assertEqual(
            ctx.exception.args,
            (
                "Creating a ModelForm without either the 'fields' attribute or the 'exclude' attribute is prohibited; "
                "form OwnerForm needs updating.",
            ),
        )

    def test_modelform_no_model(self):
        class OwnerForm(forms.ModelForm):
            pass

        with self.assertRaises(ValueError) as ctx:
            OwnerForm()

        self.assertEqual(ctx.exception.args, ("ModelForm has no model class specified.",))

    def test_modelform_no_session(self):
        class OwnerForm(forms.ModelForm):
            class Meta:
                model = Owner
                fields = forms.ALL_FIELDS

        with self.assertRaises(ValueError) as ctx:
            OwnerForm()

        self.assertEqual(ctx.exception.args, ("ModelForm has no session specified.",))

    def test_modelform_custom_setter(self):
        class OwnerForm(forms.ModelForm):
            class Meta:
                model = Owner
                session = db
                fields = ("first_name",)

            def set_first_name(self, instance, name, field, value):
                instance.first_name = "other"

        form = OwnerForm(data={"first_name": "something"})

        instance = form.save()

        self.assertEqual(instance.first_name, "other")

    def test_modelform_validation_with_model_clean(self):
        class VehicleForm(forms.ModelForm):
            class Meta:
                model = Vehicle
                session = db
                fields = ("paint", "name")

        form = VehicleForm(data={"name": "Bad Vehicle"})
        self.assertFalse(form.is_valid())
        self.assertDictEqual(form.errors, {"__all__": ["Name cannot be `Bad Value`"]})

    def test_modelform_validation_with_field_clean(self):
        class VehicleForm(forms.ModelForm):
            class Meta:
                model = Vehicle
                session = db
                fields = ("paint", "name")

        form = VehicleForm(data={"name": "Vehicle", "paint": "pink"})
        self.assertFalse(form.is_valid())
        self.assertDictEqual(form.errors, {"paint": ["Can't have a pink car"]})

    def test_modelform_save_with_errors(self):
        class VehicleForm(forms.ModelForm):
            class Meta:
                model = Vehicle
                session = db
                fields = ("type", "name")

        form = VehicleForm(data={"name": "something"})

        self.assertFalse(form.is_valid())
        with self.assertRaises(ValueError) as ctx:
            form.save()

        self.assertEqual(ctx.exception.args, ("The Vehicle could not be saved because the data didn't validate.",))

    def test_modelform_model_full_clean_validate(self):
        class BadModelForm(forms.ModelForm):
            class Meta:
                model = ModelFullCleanFail
                session = db
                fields = ("name",)

        form = BadModelForm(data={"name": "bad"})
        self.assertFalse(form.is_valid())
        self.assertDictEqual(form.errors, {"__all__": ["bad model"]})

    def test_modelform_clean_with_classic_model(self):
        class ClassicModelForm(forms.ModelForm):
            class Meta:
                model = ClassicModel
                session = db
                fields = ("name",)

        form = ClassicModelForm(data={"name": "classic"})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.instance.name, "classic")

    def test_modelform_factory_with_formfield_callback(self):
        self.callback_called = False

        def callback(*args, **kwargs):
            self.callback_called = True

        forms.modelform_factory(Owner, fields=forms.ALL_FIELDS, formfield_callback=callback, session=db)

        self.assertTrue(self.callback_called)

    def test_modelform_factory_with_no_fields_exclude(self):
        with self.assertRaises(ImproperlyConfigured) as ctx:
            forms.modelform_factory(Owner, session=db)

        self.assertEqual(
            ctx.exception.args,
            ("Calling modelform_factory without defining 'fields' or 'exclude' explicitly is prohibited.",),
        )

    def test_update_errors(self):
        class OwnerForm(forms.ModelForm):
            class Meta:
                model = Owner
                session = db
                fields = ("first_name", "last_name")
                error_messages = {"__all__": {"required": "Last name is required"}}

        form = OwnerForm(data={"first_name": "name"})
        self.assertTrue(form.is_valid())

        form = OwnerForm(data={"first_name": "name"})
        form._update_errors("error")
        self.assertDictEqual(form.errors, {"__all__": ["error"]})

        form = OwnerForm(data={"first_name": "name"})
        form._update_errors({"first_name": "name error"})
        self.assertDictEqual(form.errors, {"first_name": ["name error"]})

        form = OwnerForm(data={"first_name": "name"})
        error = ValidationError({"first_name": "name error"})
        form._update_errors(error)
        self.assertDictEqual(form.errors, {"first_name": ["name error"]})

        form = OwnerForm(data={"first_name": "name"})
        error = ValidationError({"dummy": ValidationError("error")})
        with self.assertRaises(ValueError):
            form._update_errors(error)

        form = OwnerForm(data={"first_name": "name"})
        error = ValidationError({"first_name": [ValidationError("first_name", code="required")]})
        form._update_errors(error)
        self.assertDictEqual(form.errors, {"first_name": ["This field is required."]})

        form = OwnerForm(data={"last_name": "name"})
        error = ValidationError(ValidationError("Last name is required."))
        form._update_errors(error)
        self.assertDictEqual(form.errors, {"__all__": ["Last name is required."]})



from django_sorcery.db import databases


db = databases.get("minimal")


class Asset(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))


class OrderItem(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))


class Customer(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))


class Profile(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=20))
    customer = db.OneToOne(Customer, backref="profile")


class Contact(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))


class Order(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))

    order_item = db.ManyToOne(OrderItem, backref=db.backref("orders"), fk_name="fk_order_order_item")

    applicant = db.ManyToOne(Customer, backref=db.backref("orders"), fk_name="fk_order_applicant")

    coapplicant = db.ManyToOne(Customer, backref=db.backref("other_orders"), fk_name="fk_order_coapplicant")

    assets = db.OneToMany(Asset, backref="order", fk_name="fk_order_assets")

    contacts = db.ManyToMany(Contact, backref="orders", table_name="order_contacts")


db.configure_mappers()
db.create_all()



"""Validation runner."""

from django.core.exceptions import NON_FIELD_ERRORS, ValidationError

from ..exceptions import NestedValidationError


class ValidationRunner:
    """Helper class that can execute a bunch of validators on a given
    object."""

    def __init__(self, name=None, validators=None, errors=None):
        self.name = name or NON_FIELD_ERRORS
        self.validators = validators or []
        self.errors = errors or {}

    def is_valid(self, value, raise_exception=False):
        """Runs all validators on given value and collect validation errors."""
        for v in self.validators:
            try:
                v(value)
            except ValidationError as e:
                e = NestedValidationError(e)
                try:
                    e.error_list
                except AttributeError:
                    self.errors = e.update_error_dict(self.errors)
                else:
                    self.errors.setdefault(self.name, []).extend(e.error_list)

        if self.errors and raise_exception:
            raise NestedValidationError(self.errors)

        return not self.errors



from django.core.exceptions import ValidationError
from django_sorcery.db import meta  # noqa

from ...base import TestCase
from ...testapp.models import Address, Business, Point, Vertex


class TestCompositeMeta(TestCase):
    def test_composite_meta(self):
        info = meta.model_info(Vertex)
        self.assertEqual(set(info.composites.keys()), {"start", "end"})

        self.assertListEqual(
            repr(info).split("\n"),
            [
                "<model_info(Vertex)>",
                "    <integer_column_info(Vertex.pk) pk>",
                "    <composite_info(Point, Vertex.end)>",
                "        <integer_column_info(Point.x)>",
                "        <integer_column_info(Point.y)>",
                "    <composite_info(Point, Vertex.start)>",
                "        <integer_column_info(Point.x)>",
                "        <integer_column_info(Point.y)>",
            ],
        )

        start = info.composites["start"]
        self.assertEqual(set(start.properties.keys()), {"x", "y"})
        self.assertEqual(start.properties["x"].property, Vertex.x1.property)
        self.assertEqual(start.properties["y"].property, Vertex.y1.property)
        self.assertIs(start.attribute, Vertex.start)

        end = info.composites["end"]
        self.assertEqual(set(end.properties.keys()), {"x", "y"})
        self.assertEqual(end.properties["x"].property, Vertex.x2.property)
        self.assertEqual(end.properties["y"].property, Vertex.y2.property)
        self.assertIs(end.attribute, Vertex.end)

        self.assertEqual(set(start.field_names), {"x", "y"})
        self.assertEqual(start.model_class, Point)

    def test_full_clean(self):
        instance = Business(name="foo")
        instance.other_location = Address(street="a", state="ny", zip="0456a")

        with self.assertRaises(ValidationError) as e:
            instance.full_clean()

        self.assertEqual(
            e.exception.message_dict,
            {
                "location": ["Primary location is required when other location is provided."],
                "other_location": {
                    "state": ["Select a valid choice. ny is not one of the available choices."],
                    "street": ["Street should be at least 2 characters."],
                    "zip": ["Zip cannot start with 0."],
                },
            },
        )

    def test_full_clean_exclude_inner_composite_fields(self):
        instance = Business(name="foo")
        instance.other_location = Address(street="a", state="ny", zip="0456a")

        with self.assertRaises(ValidationError) as e:
            instance.full_clean(exclude={"other_location": ["zip"]})

        self.assertEqual(
            e.exception.message_dict,
            {
                "location": ["Primary location is required when other location is provided."],
                "other_location": {
                    "street": ["Street should be at least 2 characters."],
                    "state": ["Select a valid choice. ny is not one of the available choices."],
                },
            },
        )

    def test_full_clean_exclude_complete_composite(self):
        instance = Business(name="foo")
        instance.location = Address(street="a", state=None, zip=None)
        instance.other_location = Address(street="a", state="ny", zip="0456a")

        with self.assertRaises(ValidationError) as e:
            instance.full_clean(exclude={"other_location": None})

        self.assertEqual(
            e.exception.message_dict,
            {
                "location": {
                    "__all__": ["All state, street, zip are required."],
                    "street": ["Street should be at least 2 characters."],
                }
            },
        )



"""Django-esque declarative fields for sqlalchemy."""
from contextlib import suppress

import sqlalchemy as sa
from django import forms as djangoforms
from django.core import validators as django_validators
from django.db.backends.base import operations
from django.forms import fields as djangofields
from django.utils.module_loading import import_string

from .url import DIALECT_MAP_TO_DJANGO


__all__ = [
    "BigIntegerField",
    "BinaryField",
    "BooleanField",
    "CharField",
    "DateField",
    "DateTimeField",
    "DecimalField",
    "DurationField",
    "EmailField",
    "EnumField",
    "Field",
    "FloatField",
    "IntegerField",
    "NullBooleanField",
    "SlugField",
    "SmallIntegerField",
    "TextField",
    "TimeField",
    "TimestampField",
    "URLField",
]


class Field(sa.Column):
    """Base django-esque field."""

    default_validators = []
    type_class = None
    form_class = None
    widget_class = None

    def __init__(self, *args, **kwargs):
        self.db = kwargs.pop("db", None)

        name = None
        args = list(args)
        if args and isinstance(args[0], str):
            name = args.pop(0)

        column_type = kwargs.pop("type_", None)
        if args and hasattr(args[0], "_sqla_type"):
            column_type = args.pop(0)

        validators = kwargs.pop("validators", [])
        required = kwargs.pop("required", None)
        column_kwargs = self.get_column_kwargs(kwargs)

        if column_type is None:
            type_class = self.get_type_class(kwargs)
            type_kwargs = self.get_type_kwargs(type_class, kwargs)
            column_type = self.get_type(type_class, type_kwargs)

        column_kwargs["type_"] = column_type
        column_kwargs.setdefault("name", name)
        super().__init__(*args, **column_kwargs)
        self.info["validators"] = self.get_validators(validators)
        self.info["required"] = required if required is not None else not self.nullable
        label = kwargs.pop("label", None)
        if label:
            self.info["label"] = label
        self.info["form_class"] = kwargs.pop("form_class", None) or self.get_form_class(kwargs)
        if self.widget_class:
            self.info["widget_class"] = self.widget_class

    def get_form_class(self, kwargs):
        """Returns form field class."""
        return self.form_class

    def get_type_kwargs(self, type_class, kwargs):
        """Returns sqlalchemy type kwargs."""
        type_args = sa.util.get_cls_kwargs(type_class)
        return {k: kwargs.pop(k) for k in type_args if not k.startswith("*") and k in kwargs}

    def get_column_kwargs(self, kwargs):
        """Returns sqlalchemy column kwargs."""
        column_args = [
            "autoincrement",
            "comment",
            "default",
            "doc",
            "index",
            "info",
            "key",
            "name",
            "onupdate",
            "server_default",
            "server_onupdate",
            "type_",
            "unique",
            "_proxies",
        ]
        column_kwargs = {k: kwargs.pop(k) for k in column_args if k in kwargs}
        column_kwargs["primary_key"] = kwargs.pop("primary_key", False)
        column_kwargs["nullable"] = kwargs.pop("nullable", not column_kwargs["primary_key"])
        return column_kwargs

    def get_type_class(self, kwargs):
        """Returns sqlalchemy column type."""
        return self.type_class

    def get_validators(self, validators):
        """Returns django validators for the field."""
        return self.default_validators[:] + validators

    def get_type(self, type_class, type_kwargs):
        """Returns sqlalchemy column type instance for the field."""
        return type_class(**type_kwargs)


class BooleanField(Field):
    """Django like boolean field."""

    type_class = sa.Boolean
    form_class = djangofields.BooleanField

    def get_type_kwargs(self, type_class, kwargs):
        type_kwargs = super().get_type_kwargs(type_class, kwargs)
        type_kwargs["name"] = kwargs.pop("constraint_name", None)
        return type_kwargs

    def get_column_kwargs(self, kwargs):
        column_kwargs = super().get_column_kwargs(kwargs)
        column_kwargs["nullable"] = False
        column_kwargs.setdefault("default", False)
        return column_kwargs


class CharField(Field):
    """Django like char field."""

    type_class = sa.String
    length_is_required = True
    form_class = djangofields.CharField

    def get_type_kwargs(self, type_class, kwargs):
        type_kwargs = super().get_type_kwargs(type_class, kwargs)
        type_kwargs["length"] = type_kwargs.get("length") or kwargs.get("max_length")
        if not type_kwargs["length"] and self.length_is_required:
            raise TypeError('Missing length parameter. Must provide either "max_length" or "length" parameter')
        return type_kwargs

    def get_validators(self, validators):
        validators = super().get_validators(validators)
        if self.type.length and not any(isinstance(i, django_validators.MaxLengthValidator) for i in validators):
            validators.append(django_validators.MaxLengthValidator(self.type.length))
        return validators


class DateField(Field):
    """Django like date field."""

    type_class = sa.Date
    form_class = djangofields.DateField


class DateTimeField(Field):
    """Django like datetime field."""

    type_class = sa.DateTime
    form_class = djangofields.DateTimeField


class DurationField(Field):
    """Django like duration field."""

    type_class = sa.Interval
    form_class = djangofields.DurationField


class DecimalField(Field):
    """Django like decimal field."""

    type_class = sa.Numeric
    form_class = djangofields.DecimalField

    def get_type_kwargs(self, type_class, kwargs):
        type_kwargs = super().get_type_kwargs(type_class, kwargs)
        type_kwargs.setdefault("precision", kwargs.pop("max_digits", None))
        type_kwargs.setdefault("scale", kwargs.pop("decimal_places", None))
        type_kwargs["asdecimal"] = True
        return type_kwargs

    def get_validators(self, validators):
        return super().get_validators(validators) + [
            django_validators.DecimalValidator(self.type.precision, self.type.scale)
        ]


class EmailField(CharField):
    """Django like email field."""

    default_validators = [django_validators.validate_email]
    form_class = djangofields.EmailField


class EnumField(Field):
    """Django like choice field that uses an enum sqlalchemy type."""

    type_class = sa.Enum

    def get_type_kwargs(self, type_class, kwargs):
        type_kwargs = super().get_type_kwargs(type_class, kwargs)
        choices = kwargs.pop("choices", None) or kwargs.pop("enum_class", None)
        type_kwargs["choices"] = choices if len(choices) > 1 and isinstance(choices, (list, tuple)) else [choices]
        type_kwargs["name"] = kwargs.pop("constraint_name", None)

        enum_args = [
            "native_enum",
            "create_constraint",
            "values_callable",
            "convert_unicode",
            "validate_strings",
            "schema",
            "quote",
            "_create_events",
        ]
        for k in enum_args:
            if k in kwargs:
                type_kwargs[k] = kwargs.pop(k)

        return type_kwargs

    def get_type(self, type_class, type_kwargs):
        choices = type_kwargs.pop("choices")
        return type_class(*choices, **type_kwargs)

    def get_form_class(self, kwargs):
        if self.type.enum_class:
            from ..fields import EnumField as EnumFormField

            return EnumFormField

        return djangofields.TypedChoiceField


class FloatField(Field):
    """Django like float field."""

    type_class = sa.Float
    form_class = djangofields.FloatField

    def get_type_kwargs(self, type_class, kwargs):
        type_kwargs = super().get_type_kwargs(type_class, kwargs)
        type_kwargs["precision"] = type_kwargs.get("precision") or kwargs.pop("max_digits", None)
        return type_kwargs


class ValidateIntegerFieldMixin:
    """A mixin that provides default min/max validators for integer types."""

    def get_django_dialect_ranges(self):
        """Returns django min/max ranges using current dialect."""
        ops = operations.BaseDatabaseOperations
        with suppress(ImportError):
            ops = (
                import_string(DIALECT_MAP_TO_DJANGO.get(self.db.url.get_dialect().name) + ".base.DatabaseOperations")
                if self.db
                else operations.BaseDatabaseOperations
            )

        return ops.integer_field_ranges

    def get_dialect_range(self):
        """Returns the min/max ranges supported by dialect."""
        return self.get_django_dialect_ranges()[self.__class__.__name__]

    def get_validators(self, validators):
        """Returns django integer min/max validators supported by the
        database."""
        validators = super().get_validators(validators)
        min_int, max_int = self.get_dialect_range()
        if not any(isinstance(i, django_validators.MinValueValidator) for i in validators):
            validators.append(django_validators.MinValueValidator(min_int))
        if not any(isinstance(i, django_validators.MaxValueValidator) for i in validators):
            validators.append(django_validators.MaxValueValidator(max_int))
        return validators


class IntegerField(ValidateIntegerFieldMixin, Field):
    """Django like integer field."""

    default_validators = [django_validators.validate_integer]
    type_class = sa.Integer
    form_class = djangofields.IntegerField


class BigIntegerField(ValidateIntegerFieldMixin, Field):
    """Django like big integer field."""

    default_validators = [django_validators.validate_integer]
    type_class = sa.BigInteger
    form_class = djangofields.IntegerField


class SmallIntegerField(ValidateIntegerFieldMixin, Field):
    """Django like small integer field."""

    default_validators = [django_validators.validate_integer]
    type_class = sa.SmallInteger
    form_class = djangofields.IntegerField


class NullBooleanField(BooleanField):
    """Django like nullable boolean field."""

    form_class = djangofields.NullBooleanField

    def get_column_kwargs(self, kwargs):
        kwargs["nullable"] = True
        return kwargs


class SlugField(CharField):
    """Django like slug field."""

    default_validators = [django_validators.validate_slug]
    form_class = djangofields.SlugField


class TextField(CharField):
    """Django like text field."""

    type_class = sa.Text
    length_is_required = False
    form_class = djangofields.CharField
    widget_class = djangoforms.Textarea


class TimeField(Field):
    """Django like time field."""

    type_class = sa.Time
    form_class = djangofields.TimeField


class TimestampField(DateTimeField):
    """Django like datetime field that uses timestamp sqlalchemy type."""

    type_class = sa.TIMESTAMP


class URLField(CharField):
    """Django like url field."""

    default_validators = [django_validators.URLValidator()]
    form_class = djangofields.URLField


class BinaryField(Field):
    """Django like binary field."""

    type_class = sa.LargeBinary
    length_is_required = False



from collections import namedtuple

from django.core.exceptions import ValidationError
from django_sorcery.exceptions import NestedValidationError
from django_sorcery.validators import (
    ValidateCantRemove,
    ValidateEmptyWhen,
    ValidateNotEmptyWhen,
    ValidateOnlyOneOf,
    ValidateTogetherModelFields,
    ValidateValue,
    ValidationRunner,
)

from .base import TestCase
from .testapp.models import ValidateUniqueModel, db


class TestValidatorRunner(TestCase):
    def test_is_valid(self):
        runner = ValidationRunner(validators=[bool])
        self.assertTrue(runner.is_valid(1))

    def test_invalid(self):
        def invalid(x):
            raise ValidationError("invalid")

        runner = ValidationRunner(validators=[invalid])
        self.assertFalse(runner.is_valid(1))

    def test_invalid_nested(self):
        def invalid(x):
            raise NestedValidationError({"bad": ["really"]})

        runner = ValidationRunner(validators=[invalid])
        self.assertFalse(runner.is_valid(1))
        self.assertEqual(len(runner.errors), 1)
        self.assertEqual(len(runner.errors["bad"]), 1)
        self.assertIsInstance(runner.errors["bad"][0], NestedValidationError)

    def test_invalid_raise_error(self):
        def invalid(x):
            raise ValidationError("invalid")

        runner = ValidationRunner(validators=[invalid])
        with self.assertRaises(ValidationError):
            runner.is_valid(1, raise_exception=True)


Node = namedtuple("Node", ["left", "right", "value"])
Node.__new__.__defaults__ = (None, None, None)


class TestValidateTogetherModelFields(TestCase):
    def test_success_all_none(self):
        node = Node()

        validator = ValidateTogetherModelFields(["left", "right", "value"])

        self.assertIsNone(validator(node))

    def test_success_all_full(self):
        node = Node("left", "right", "value")

        validator = ValidateTogetherModelFields(["left", "right", "value"])

        self.assertIsNone(validator(node))

    def test_success_fail(self):
        node = Node("left", "right")

        validator = ValidateTogetherModelFields(["left", "right", "value"])

        with self.assertRaises(ValidationError):
            self.assertIsNone(validator(node))


class TestValidateUnique(TestCase):
    def test_validate_on_insert(self):
        model = ValidateUniqueModel(name="Test")
        db.add(model)
        db.flush()

        model = ValidateUniqueModel(name="Test")
        db.add(model)
        with self.assertRaises(ValidationError) as ctx:
            db.flush()

        self.assertEqual(
            ctx.exception.message_dict,
            {"__all__": ["name must make a unique set.", "bar, foo must make a unique set."]},
        )

    def test_validate_on_update(self):
        model = ValidateUniqueModel(name="Name", foo="Test", bar="Test")
        db.add(model)
        db.flush()

        model.foo = "foo"
        db.flush()


class TestValidateValue(TestCase):
    def setUp(self):
        super().setUp()
        self.validator = ValidateValue("right", lambda n: n.left != "left" or "right" in n.right)

    def test_valid(self):
        self.assertIsNone(self.validator(Node("left", "right")))
        self.assertIsNone(self.validator(Node("notleft", "notRight")))
        self.assertIsNone(self.validator(Node("notleft", "right")))

    def test_invalid(self):
        with self.assertRaises(ValidationError):
            self.validator(Node("left", "left"))

    def test_error(self):
        with self.assertRaises(ValidationError):
            self.validator(Node("left", None))


class TestValidateEmptyWhen(TestCase):
    def setUp(self):
        super().setUp()
        self.validator = ValidateEmptyWhen("right", lambda n: not n.left)

    def test_valid(self):
        self.assertIsNone(self.validator(Node(None, None)))
        self.assertIsNone(self.validator(Node("left", "right")))

    def test_invalid(self):
        with self.assertRaises(ValidationError):
            self.validator(Node(None, "right"))


class TestValidateNotEmptyWhen(TestCase):
    def setUp(self):
        super().setUp()
        self.validator = ValidateNotEmptyWhen("right", lambda n: n.left)

    def test_valid(self):
        self.assertIsNone(self.validator(Node(None, None)))
        self.assertIsNone(self.validator(Node("left", "right")))

    def test_invalid(self):
        with self.assertRaises(ValidationError):
            self.validator(Node("left", None))


class TestValidateOnlyOneOf(TestCase):
    def test_valid_required(self):
        validator = ValidateOnlyOneOf(["left", "right"])

        self.assertIsNone(validator(Node(None, "right")))
        self.assertIsNone(validator(Node("left", None)))

    def test_invalid_required(self):
        validator = ValidateOnlyOneOf(["left", "right"])

        with self.assertRaises(ValidationError):
            validator(Node("left", "right"))
        with self.assertRaises(ValidationError):
            validator(Node(None, None))

    def test_valid_not_required(self):
        validator = ValidateOnlyOneOf(["left", "right"], required=False)

        self.assertIsNone(validator(Node(None, "right")))
        self.assertIsNone(validator(Node("left", None)))
        self.assertIsNone(validator(Node(None, None)))

    def test_invalid_not_required(self):
        validator = ValidateOnlyOneOf(["left", "right"], required=False)

        with self.assertRaises(ValidationError):
            validator(Node("left", "right"))


class TestValidateCantRemove(TestCase):
    def setUp(self):
        super().setUp()
        self.validator = ValidateCantRemove("name")

    def test_no_changes(self):
        m = ValidateUniqueModel(name="Test")
        db.add(m)
        db.flush()

        self.assertIsNone(self.validator(m))

    def test_not_removed(self):
        m = ValidateUniqueModel(name="Test")
        db.add(m)
        db.flush()
        m.name = "Test2"

        self.assertIsNone(self.validator(m))

    def test_removed(self):
        m = ValidateUniqueModel(name="Test")
        db.add(m)
        db.flush()
        m.name = None

        with self.assertRaises(ValidationError):
            self.validator(m)



from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.utils.html import escape
from django_sorcery.views.detail import (
    SingleObjectMixin,
    SingleObjectTemplateResponseMixin,
)

from ..base import TestCase
from ..testapp.models import Owner, db


class TestDetailView(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all(
            [
                Owner(id=1, first_name="Test 1", last_name="Owner1"),
                Owner(id=2, first_name="Test 2", last_name="Owner2"),
                Owner(id=3, first_name="Test 3", last_name="Owner3"),
                Owner(id=4, first_name="Test 4", last_name="Owner4"),
            ]
        )
        db.flush()

    def test_get_object_fail_no_kwargs(self):
        class DummyView(SingleObjectMixin):
            model = Owner

        view = DummyView()
        view.kwargs = {}

        with self.assertRaises(AttributeError) as ctx:
            view.get_object(queryset=[Owner()])

        self.assertEqual(
            str(ctx.exception),
            "Generic detail view DummyView must be called with either an object pk or a slug in the URLconf.",
        )

    def test_detail_generic(self):
        url = reverse("owner_detail", kwargs={"id": 1})

        response = self.client.get(url)

        self.assertEqual(
            response.content.decode().strip(),
            f"owner{escape(repr(Owner.query.get(1)))}",
        )

    def test_detail_generic_with_template(self):
        url = reverse("owner_detail_tmpl", kwargs={"id": 1})

        response = self.client.get(url)

        self.assertEqual(response.content.decode().strip(), escape(repr(Owner.query.get(1))))

    def test_detail_generic_with_slug(self):
        url = reverse("owner_detail_slug", kwargs={"slug": "Owner1"})

        response = self.client.get(url)

        self.assertEqual(
            response.content.decode().strip(),
            f"owner{escape(repr(Owner.query.get(1)))}",
        )

    def test_detail_generic_404(self):
        url = reverse("owner_detail", kwargs={"id": 99})

        response = self.client.get(url)

        self.assertEqual(response.status_code, 404)

    def test_detail_generic_context_name(self):
        url = reverse("owner_detail_context_name", kwargs={"id": 1})

        response = self.client.get(url)

        self.assertIn("item", response.context_data)
        self.assertIsInstance(response.context_data["item"], Owner)

    def test_detail_get_template_name_field_from_object(self):
        view = SingleObjectTemplateResponseMixin()
        view.object = Owner(last_name="owner_template")
        view.template_name_field = "last_name"

        template_names = view.get_template_names()

        self.assertEqual(template_names, [view.object.last_name])

    def test_detail_get_template_name_fail(self):
        view = SingleObjectTemplateResponseMixin()
        view.object = Owner(last_name="owner_template")

        with self.assertRaises(ImproperlyConfigured):
            view.get_template_names()



"""CreateAll command."""

from django.core.management.base import BaseCommand
from sqlalchemy.orm import configure_mappers

from ...db import databases


class CreateAll(BaseCommand):
    """Creates db schema using metadata.create_all."""

    help = "Creates SQLAlchemy database schemas"

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            "-d",
            action="append",
            dest="databases",
            type=str,
            help="Nominates a database to synchronize. By default will synchronize all.",
        )

    def handle(self, *args, **kwargs):
        configure_mappers()
        dbs = kwargs.get("databases") or databases.keys()
        for key in dbs:
            databases[key].create_all()
            self.stdout.write(self.style.SUCCESS('Successfully ran create_all() for "%s"' % key))


Command = CreateAll



import datetime
from decimal import Decimal

from django.core.exceptions import ValidationError
from django_sorcery.db import meta, models
from django_sorcery.utils import make_args

from ..base import TestCase
from ..testapp.models import (
    Address,
    AllKindsOfFields,
    Business,
    DummyEnum,
    Option,
    Owner,
    Part,
    SelectedAutoCoerce,
    Vehicle,
    VehicleType,
    db,
)


class TestModelRepr(TestCase):
    def test_simple_repr(self):
        vehicle = Vehicle()
        self.assertEqual(models.simple_repr(vehicle), "Vehicle(id=None, is_used=False)")

        vehicle.name = "Test"
        self.assertEqual(models.simple_repr(vehicle), "Vehicle(id=None, is_used=False, name='Test')")

        vehicle.id = 1234
        self.assertTrue(models.simple_repr(vehicle), "Vehicle(id=1234, is_used=False, name='Test')")

        vehicle.id = "abc"
        self.assertTrue(models.simple_repr(vehicle), "Vehicle(id='abc', is_used=False, name='Test')")

        vehicle.id = b"abc"
        self.assertTrue(models.simple_repr(vehicle), "Vehicle(id='abc', is_used=False, name='Test')")


class TestSerialization(TestCase):
    def test_serialize_none(self):
        self.assertIsNone(models.serialize(None))

    def test_shallow_serialize(self):
        vehicle = Vehicle(owner=Owner(first_name="first_name", last_name="last_name"), type=VehicleType.car)

        self.assertDictEqual(
            {
                "_owner_id": None,
                "created_at": None,
                "id": None,
                "is_used": False,
                "name": None,
                "msrp": None,
                "paint": None,
                "type": VehicleType.car,
            },
            models.serialize(vehicle),
        )

    def test_serialize_with_composites(self):
        business = Business(
            name="test",
            location=Address(street="street 1", state="state 1", zip="zip 1"),
            other_location=Address(street="street 2", state="state 2", zip="zip 2"),
        )

        self.assertDictEqual(
            {
                "id": None,
                "name": "test",
                "employees": 5,
                "location": {"state": "state 1", "street": "street 1", "zip": "zip 1"},
                "other_location": {"state": "state 2", "street": "street 2", "zip": "zip 2"},
            },
            models.serialize(business),
        )

    def test_serialize_with_relations(self):
        vehicle = Vehicle(
            name="vehicle",
            owner=Owner(first_name="first_name", last_name="last_name"),
            is_used=True,
            paint="red",
            type=VehicleType.car,
            options=[Option(name="option 1"), Option(name="option 2")],
            parts=[Part(name="part 1"), Part(name="part 2")],
        )

        self.assertDictEqual(
            {
                "_owner_id": None,
                "created_at": None,
                "id": None,
                "is_used": True,
                "msrp": None,
                "name": "vehicle",
                "options": [{"id": None, "name": "option 1"}, {"id": None, "name": "option 2"}],
                "owner": {"id": None, "first_name": "first_name", "last_name": "last_name"},
                "paint": "red",
                "parts": [{"id": None, "name": "part 1"}, {"id": None, "name": "part 2"}],
                "type": VehicleType.car,
            },
            models.serialize(vehicle, Vehicle.owner, Vehicle.options, Vehicle.parts),
        )

    def test_deserialize(self):
        data = {
            "_owner_id": None,
            "created_at": None,
            "id": 1,
            "is_used": True,
            "name": "vehicle",
            "options": [{"id": 3, "name": "option 1"}, {"id": 4, "name": "option 2"}],
            "owner": {"id": 2, "first_name": "first_name", "last_name": "last_name"},
            "paint": "red",
            "parts": [{"id": 5, "name": "part 1"}, {"id": 6, "name": "part 2"}],
            "type": VehicleType.car,
        }

        vehicle = models.deserialize(Vehicle, data)

        self.assertDictEqual(
            {
                "_owner_id": None,
                "created_at": None,
                "id": 1,
                "is_used": True,
                "msrp": None,
                "name": "vehicle",
                "options": [{"id": 3, "name": "option 1"}, {"id": 4, "name": "option 2"}],
                "owner": {"id": 2, "first_name": "first_name", "last_name": "last_name"},
                "paint": "red",
                "parts": [{"id": 5, "name": "part 1"}, {"id": 6, "name": "part 2"}],
                "type": VehicleType.car,
            },
            models.serialize(vehicle, Vehicle.owner, Vehicle.options, Vehicle.parts),
        )

    def test_deserialize_none(self):
        self.assertIsNone(models.deserialize(Vehicle, None))

    def test_deserialize_with_map(self):
        data = [
            {
                "_owner_id": 2,
                "created_at": None,
                "id": 1,
                "is_used": True,
                "paint": "red",
                "type": VehicleType.car,
                "name": "vehicle",
                "owner": {"id": 2, "first_name": "first_name", "last_name": "last_name"},
                "options": [{"id": 3, "name": "option 1"}, {"id": 4, "name": "option 2"}],
                "parts": [{"id": 5, "name": "part 1"}, {"id": 6, "name": "part 2"}],
            },
            {
                "_owner_id": 2,
                "created_at": None,
                "id": 7,
                "is_used": True,
                "paint": "red",
                "type": VehicleType.car,
                "name": "vehicle",
                # missing owner to test back population of many to ones
                "options": [{"id": 3, "name": "option 1"}, {"id": 4, "name": "option 2"}],
                "parts": [{"id": 5, "name": "part 1"}, {"id": 6, "name": "part 2"}],
            },
        ]

        vehicle1, vehicle2 = models.deserialize(Vehicle, data)

        self.assertEqual(vehicle1.owner, vehicle2.owner)

    def test_deserialize_composites(self):
        data = {
            "id": None,
            "name": "test",
            "location": {"state": "state 1", "street": "street 1", "zip": "zip 1"},
            "other_location": {"state": "state 2", "street": "street 2", "zip": "zip 2"},
        }

        business = models.deserialize(Business, data)

        self.assertDictEqual(
            {
                "id": None,
                "name": "test",
                "employees": 5,
                "location": {"state": "state 1", "street": "street 1", "zip": "zip 1"},
                "other_location": {"state": "state 2", "street": "street 2", "zip": "zip 2"},
            },
            models.serialize(business),
        )


class TestClone(TestCase):
    def setUp(self):
        super().setUp()
        self.vehicle = Vehicle(
            name="vehicle",
            owner=Owner(first_name="first_name", last_name="last_name"),
            is_used=True,
            paint="red",
            type=VehicleType.car,
            options=[Option(name="option 1"), Option(name="option 2")],
            parts=[Part(name="part 1"), Part(name="part 2")],
        )

        db.add(self.vehicle)
        db.flush()
        db.expire_all()

    def test_clone_none(self):
        self.assertIsNone(models.clone(None))

    def test_shallow_clone(self):
        clone = models.clone(self.vehicle)
        db.add(clone)
        db.flush()

        self.assertNotEqual(clone, self.vehicle)
        self.assertNotEqual(clone.as_dict(), self.vehicle.as_dict())
        self.assertNotEqual(clone.id, self.vehicle.id)
        self.assertEqual(clone.is_used, self.vehicle.is_used)
        self.assertEqual(clone.msrp, self.vehicle.msrp)
        self.assertEqual(clone.name, self.vehicle.name)
        self.assertEqual(clone.paint, self.vehicle.paint)
        self.assertEqual(clone.type, self.vehicle.type)
        self.assertIsNone(clone.owner)
        self.assertEqual(clone.options, [])
        self.assertEqual(clone.parts, [])

    def test_clone_with_relation(self):
        clone = models.clone(
            self.vehicle, Vehicle.owner, paint="blue", options=self.vehicle.options, parts=self.vehicle.parts
        )
        db.add(clone)
        db.flush()

        self.assertNotEqual(clone, self.vehicle)
        self.assertNotEqual(clone.as_dict(), self.vehicle.as_dict())
        self.assertNotEqual(clone.id, self.vehicle.id)
        self.assertEqual(clone.paint, "blue")
        self.assertEqual(clone.name, self.vehicle.name)
        clone.paint = "red"

        self.assertIsNot(clone.owner, self.vehicle.owner)
        self.assertNotEqual(clone.owner.as_dict(), self.vehicle.owner.as_dict())
        self.assertNotEqual(clone.owner.id, self.vehicle.owner.id)
        self.assertEqual(clone.owner.first_name, self.vehicle.owner.first_name)
        self.assertEqual(clone.owner.last_name, self.vehicle.owner.last_name)

        self.assertEqual(clone.options, self.vehicle.options)
        self.assertEqual(clone.parts, self.vehicle.parts)

    def test_clone_with_composite(self):
        business = Business(
            name="test",
            location=Address(street="street 1", state="state 1", zip="zip 1"),
            other_location=Address(street="street 2", state="state 2", zip="zip 2"),
        )

        clone = models.clone(business)

        self.assertIsNot(clone, business)
        self.assertIsNot(clone.location, business.location)
        self.assertIsNot(clone.other_location, business.other_location)

        self.assertEqual(clone.location, business.location)
        self.assertEqual(clone.other_location, business.other_location)

    def test_clone_with_relation_options(self):
        clone = models.clone(self.vehicle, make_args(Vehicle.owner, first_name="test"))
        db.add(clone)
        db.flush()

        self.assertNotEqual(clone, self.vehicle)
        self.assertNotEqual(clone.as_dict(), self.vehicle.as_dict())
        self.assertNotEqual(clone.id, self.vehicle.id)

        self.assertNotEqual(clone.owner, self.vehicle.owner)
        self.assertNotEqual(clone.owner.as_dict(), self.vehicle.owner.as_dict())
        self.assertNotEqual(clone.owner.id, self.vehicle.owner.id)
        self.assertEqual(clone.owner.first_name, "test")

    def test_clone_list_relation(self):
        clone = models.clone(self.vehicle, Vehicle.options)
        db.add(clone)
        db.flush()

        self.assertNotEqual(clone, self.vehicle)
        self.assertNotEqual(clone.as_dict(), self.vehicle.as_dict())
        self.assertNotEqual(clone.id, self.vehicle.id)

        for cloned, orig in zip(clone.options, self.vehicle.options):
            self.assertNotEqual(cloned, orig)
            self.assertNotEqual(cloned.as_dict(), orig.as_dict())
            self.assertNotEqual(cloned.id, orig.id)


class TestBaseModel(TestCase):
    def test_default_table_name(self):
        self.assertEqual(AllKindsOfFields.__table__.name, "all_kinds_of_fields")

    def test_as_dict(self):
        instance = AllKindsOfFields()

        self.assertDictEqual(
            instance.as_dict(),
            {
                "bigint": None,
                "biginteger": None,
                "boolean": None,
                "boolean_notnull": None,
                "char": None,
                "date": None,
                "datetime": None,
                "decimal": None,
                "enum": None,
                "enum_choice": None,
                "float": None,
                "int": None,
                "integer": None,
                "interval": None,
                "largebinary": None,
                "nchar": None,
                "numeric": None,
                "pk": None,
                "real": None,
                "smallint": None,
                "smallinteger": None,
                "string": None,
                "text": None,
                "time": None,
                "timestamp": None,
                "unicode": None,
                "unicodetext": None,
                "varchar": None,
            },
        )

    def test_model_repr(self):
        owner = Owner(id=1, first_name="Morty", last_name="McFly")

        self.assertEqual(repr(owner), "Owner(id=1, first_name='Morty', last_name='McFly')")

    def test_get_properties_for_validation(self):
        instance = Owner(id=1, first_name="Morty", last_name="McFly")

        self.assertEqual(set(instance._get_properties_for_validation()), {"first_name", "last_name"})

    def test_get_nested_objects_for_validation(self):
        instance = Business()

        self.assertEqual(set(instance._get_nested_objects_for_validation()), {"location", "other_location"})

    def test_get_relation_objects_for_validation(self):
        instance = Vehicle()
        info = meta.model_info(instance)

        self.assertDictEqual(
            instance._get_relation_objects_for_validation(),
            {"options": info.options, "owner": info.owner, "parts": info.parts},
        )


class TestAutoCoerce(TestCase):
    def setUp(self):
        self.instance = AllKindsOfFields()

    def _run_tests(self, attr, tests):
        for test, exp in tests:
            if exp is ValidationError:
                with self.assertRaises(ValidationError):
                    setattr(self.instance, attr, test)
            else:
                setattr(self.instance, attr, test)
                self.assertEqual(getattr(self.instance, attr), exp)

    def test_boolean(self):
        self.instance.boolean_notnull = "true"
        self.assertIsInstance(self.instance.boolean_notnull, bool)

        self.instance.boolean = "true"
        self.assertTrue(self.instance.boolean)
        self.assertTrue(type(self.instance.boolean) is bool)

        self.instance.boolean = "false"
        self.assertFalse(self.instance.boolean)
        self.assertTrue(type(self.instance.boolean) is bool)

    def test_enum(self):
        self.instance.enum = 0
        self.assertTrue(self.instance.enum is DummyEnum.zero)

        self.instance.enum = "one"
        self.assertTrue(self.instance.enum is DummyEnum.one)

        self.instance.enum = 2
        self.assertTrue(self.instance.enum is DummyEnum.two)

        self.instance.enum = DummyEnum.one
        self.assertTrue(self.instance.enum is DummyEnum.one)

        with self.assertRaises(ValidationError):
            self.instance.enum = "three"

    def test_enum_choice(self):
        self.instance.enum_choice = "three"
        self.assertEqual(self.instance.enum_choice, "three")

        with self.assertRaises(ValidationError):
            self.instance.enum_choice = "one"

    def test_string(self):
        tests = [
            ("abc", "abc"),
            (1234, "1234"),
            ("", ""),
            ("é", "é"),
            (Decimal("1234.44"), "1234.44"),
            (None, None),
            # excess whitespace tests
            ("\t\t\t\t\n", ""),
            ("\t\tabc\t\t\n", "abc"),
            ("\t\t20,000\t\t\n", "20,000"),
            ("  \t 23\t", "23"),
        ]

        for attr in ["char", "nchar", "string", "text", "unicode", "unicodetext", "varchar"]:
            self._run_tests(attr, tests)

    def test_decimal(self):
        tests = [
            ("1", Decimal("1")),
            ("abc", ValidationError),
            (1, Decimal("1")),
            ("20000", Decimal("20000")),
            ("20,000", Decimal("20000")),
            ("1.e-8", Decimal("1E-8")),
            ("1.-8", ValidationError),
            ("", None),
            (None, None),
            # excess whitespace tests
            ("\t\t\t\t\n", None),
            ("\t\tabc\t\t\n", ValidationError),
            ("\t\t20,000\t\t\n", Decimal("20000")),
            ("  \t 23\t", Decimal("23")),
        ]
        for attr in ["decimal", "numeric"]:
            self._run_tests(attr, tests)

        with self.settings(THOUSAND_SEPARATOR=".", DECIMAL_SEPARATOR=","):
            tests = [("20.000", Decimal("20000")), ("20,000", Decimal("20.000"))]
            self._run_tests(attr, tests)

    def test_float(self):
        tests = [
            ("1", 1.0),
            ("abc", ValidationError),
            ("1.0", 1.0),
            ("1.", 1.0),
            ("1.001", 1.001),
            ("1.e-8", 1e-08),
            ("", None),
            (None, None),
            # excess whitespace tests
            ("\t\t\t\t\n", None),
            ("\t\tabc\t\t\n", ValidationError),
            ("\t\t20,000.02\t\t\n", 20000.02),
            ("  \t 23\t", 23.0),
        ]
        for attr in ["float", "real"]:
            self._run_tests(attr, tests)

    def test_integer(self):
        tests = [
            ("1", 1),
            ("abc", ValidationError),
            (1.0, 1),
            (1.01, ValidationError),
            ("1.0", 1),
            ("1.01", ValidationError),
            ("", None),
            (None, None),
            # # excess whitespace tests
            ("\t\t\t\t\n", None),
            ("\t\tabc\t\t\n", ValidationError),
            ("\t\t20,000.02\t\t\n", ValidationError),
            ("\t\t20,000\t\t\n", 20000),
            ("  \t 23\t", 23),
        ]
        for attr in ["int", "integer", "bigint", "biginteger", "smallint", "smallinteger"]:
            self._run_tests(attr, tests)

    def test_date(self):
        tests = [
            (datetime.date(2006, 10, 25), datetime.date(2006, 10, 25)),
            (datetime.datetime(2006, 10, 25, 14, 30), datetime.date(2006, 10, 25)),
            (datetime.datetime(2006, 10, 25, 14, 30, 59), datetime.date(2006, 10, 25)),
            (datetime.datetime(2006, 10, 25, 14, 30, 59, 200), datetime.date(2006, 10, 25)),
            ("2006-10-25", datetime.date(2006, 10, 25)),
            ("10/25/2006", datetime.date(2006, 10, 25)),
            ("10/25/06", datetime.date(2006, 10, 25)),
            ("Oct 25 2006", datetime.date(2006, 10, 25)),
            ("October 25 2006", datetime.date(2006, 10, 25)),
            ("October 25, 2006", datetime.date(2006, 10, 25)),
            ("25 October 2006", datetime.date(2006, 10, 25)),
            ("25 October, 2006", datetime.date(2006, 10, 25)),
            ("Hello", ValidationError),
        ]
        self._run_tests("date", tests)

    def test_datetime(self):
        tests = [
            ("2006-10-25 14:30:45.000200", datetime.datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("2006-10-25 14:30:45.0002", datetime.datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("2006-10-25 14:30:45", datetime.datetime(2006, 10, 25, 14, 30, 45)),
            ("2006-10-25 14:30:00", datetime.datetime(2006, 10, 25, 14, 30)),
            ("2006-10-25 14:30", datetime.datetime(2006, 10, 25, 14, 30)),
            ("2006-10-25", datetime.datetime(2006, 10, 25, 0, 0)),
            ("10/25/2006 14:30:45.000200", datetime.datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("10/25/2006 14:30:45", datetime.datetime(2006, 10, 25, 14, 30, 45)),
            ("10/25/2006 14:30:00", datetime.datetime(2006, 10, 25, 14, 30)),
            ("10/25/2006 14:30", datetime.datetime(2006, 10, 25, 14, 30)),
            ("10/25/2006", datetime.datetime(2006, 10, 25, 0, 0)),
            ("10/25/06 14:30:45.000200", datetime.datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("10/25/06 14:30:45", datetime.datetime(2006, 10, 25, 14, 30, 45)),
            ("10/25/06 14:30:00", datetime.datetime(2006, 10, 25, 14, 30)),
            ("10/25/06 14:30", datetime.datetime(2006, 10, 25, 14, 30)),
            ("10/25/06", datetime.datetime(2006, 10, 25, 0, 0)),
            ("2012-04-23T09:15:00", datetime.datetime(2012, 4, 23, 9, 15)),
            ("2012-4-9 4:8:16", datetime.datetime(2012, 4, 9, 4, 8, 16)),
            ("2012-04-23T09:15:00Z", datetime.datetime(2012, 4, 23, 9, 15, 0, 0)),
            ("2012-4-9 4:8:16-0320", datetime.datetime(2012, 4, 9, 7, 28, 16, 0)),
            ("2012-04-23T10:20:30.400+02:30", datetime.datetime(2012, 4, 23, 7, 50, 30, 400000)),
            ("2012-04-23T10:20:30.400+02", datetime.datetime(2012, 4, 23, 8, 20, 30, 400000)),
            ("2012-04-23T10:20:30.400-02", datetime.datetime(2012, 4, 23, 12, 20, 30, 400000)),
            ("Hello", ValidationError),
        ]
        self._run_tests("datetime", tests)
        self._run_tests("timestamp", tests)

    def test_interval(self):
        tests = [
            ("30", datetime.timedelta(seconds=30)),
            ("15:30", datetime.timedelta(minutes=15, seconds=30)),
            ("1:15:30", datetime.timedelta(hours=1, minutes=15, seconds=30)),
            ("1 1:15:30.3", datetime.timedelta(days=1, hours=1, minutes=15, seconds=30, milliseconds=300)),
            ("Hello", ValidationError),
        ]
        self._run_tests("interval", tests)

    def test_time(self):
        tests = [
            (datetime.time(14, 25), datetime.time(14, 25)),
            (datetime.time(14, 25, 59), datetime.time(14, 25, 59)),
            ("14:25", datetime.time(14, 25)),
            ("14:25:59", datetime.time(14, 25, 59)),
            ("Hello", ValidationError),
        ]
        self._run_tests("time", tests)

    def test_selected_fields_only(self):
        model = SelectedAutoCoerce()

        with self.assertRaises(ValidationError):
            model.foo = "abc"

        with self.assertRaises(ValidationError):
            model.bar = "abc"

        model.me = "abc"
        self.assertEqual(model.me, "abc")


class TestInstantDefaults(TestCase):
    def test(self):
        instance = Business()

        self.assertEqual(instance.employees, 5)

        instance = Business(employees=1)

        self.assertEqual(instance.employees, 1)

    def test_non_model(self):
        self.assertIsNone(models.instant_defaults(list))



from django_sorcery.db import meta  # noqa

from ...base import TestCase
from ...testapp.models import Business


class TestModelMeta(TestCase):
    def test_model_meta(self):
        info = meta.model_info(Business)

        info_from_meta = meta.model_info(Business.__mapper__)

        info_from_instance = meta.model_info(Business())

        self.assertIs(info, info_from_meta)
        self.assertIs(info_from_meta, info_from_instance)



"""Field mapping from SQLAlchemy type's to form fields."""
import json
from contextlib import suppress

from django.core.exceptions import ValidationError
from django.forms import fields as djangofields
from django.utils.translation import gettext_lazy


class EnumField(djangofields.ChoiceField):
    """Form field for using an Enum as choices."""

    empty_value = None

    def __init__(self, enum_class=None, choices=None, **kwargs):
        self.enum_class = enum_class or choices

        kwargs["choices"] = [(e.name, e.value) for e in self.enum_class]
        if not kwargs.get("required", True):
            empty_label = kwargs.pop("empty_label", "")
            kwargs["choices"] = [(self.empty_value, empty_label)] + kwargs["choices"]

        kwargs.pop("max_length", None)
        super().__init__(**kwargs)

    def to_python(self, value):
        if value in self.empty_values:
            value = None

        if value is None:
            return

        with suppress(TypeError, KeyError, ValueError):
            return self.enum_class[value]
        with suppress(TypeError, KeyError, ValueError):
            return self.enum_class(value)
        with suppress(TypeError, AttributeError):
            return getattr(self.enum_class, value)

        raise ValidationError(self.error_messages["invalid_choice"], code="invalid", params={"value": str(value)})

    def valid_value(self, value):
        return not self.required if value is None else value in self.enum_class

    def prepare_value(self, value):
        return value if value is None else value.name

    def bound_data(self, data, initial):
        value = super().bound_data(data, initial)
        return self.prepare_value(value)


class ModelChoiceIterator:
    """Iterator for sqlalchemy query for model choice fields."""

    def __init__(self, field):
        self.field = field

    def __iter__(self):
        if self.field.empty_label is not None:
            yield ("", self.field.empty_label)

        for obj in self.field.queryset:
            yield self.choice(obj)

    def __len__(self):
        return self.field.queryset.count() + (1 if self.field.empty_label is not None else 0)

    def choice(self, obj):
        """Returns choice item for django choice field."""
        return (self.field.prepare_value(obj), self.field.label_from_instance(obj))


class ModelChoiceField(djangofields.ChoiceField):
    """A ChoiceField whose choices are a sqlalchemy model relationship."""

    default_error_messages = {
        "invalid_choice": "Select a valid choice. That choice is not one of the available choices."
    }

    iterator = ModelChoiceIterator

    def __init__(
        self,
        model,
        session,
        empty_label="---------",
        required=True,
        widget=None,
        label=None,
        initial=None,
        help_text="",
        to_field_name=None,
        limit_choices_to=None,
        **kwargs,
    ):
        self.empty_label = None if required and (initial is not None) else empty_label
        djangofields.Field.__init__(
            self, required=required, widget=widget, label=label, initial=initial, help_text=help_text, **kwargs
        )

        self._choices = None
        self.model = model
        from .db import meta

        self.model_info = meta.model_info(model)
        self.session = session
        self._queryset = None
        self.limit_choices_to = limit_choices_to  # limit the queryset later.
        self.to_field_name = to_field_name

    def _get_queryset(self):
        return self._queryset or self.session.query(self.model)

    def _set_queryset(self, queryset):
        self._queryset = None if queryset is None else queryset

    queryset = property(_get_queryset, _set_queryset)

    def _get_choices(self):
        return self.iterator(self)

    choices = property(_get_choices, djangofields.ChoiceField._set_choices)

    def get_object(self, value):
        """Returns model instance."""
        if value in self.empty_values:
            return None

        pk = None
        try:
            pk = json.loads(value)
            if isinstance(pk, dict):
                pk = self.model_info.primary_keys_from_dict(pk)
        except TypeError:
            pk = value

        obj = self.session.query(self.model).get(pk)
        if obj is None:
            raise ValidationError(self.error_messages["invalid_choice"], code="invalid_choice")

        return obj

    def to_python(self, value):
        return self.get_object(value)

    def label_from_instance(self, obj):
        """Returns label from model instance."""
        return str(obj)

    def prepare_instance_value(self, obj):
        """Returns primary key from instance."""
        return self.model_info.primary_keys_from_instance(obj)

    def prepare_value(self, obj):
        if isinstance(obj, self.model):
            return self.prepare_instance_value(obj)

        return super().prepare_value(obj)

    def validate(self, value):
        return djangofields.Field.validate(self, value)

    def get_bound_field(self, form, field_name):
        self.widget.choices = self.choices
        return super().get_bound_field(form, field_name)

    def get_limit_choices_to(self):
        """Returns limit_choices_to for this model.

        If it is a callable, invoke it and return the result
        """
        if callable(self.limit_choices_to):
            return self.limit_choices_to()
        return self.limit_choices_to


class ModelMultipleChoiceField(ModelChoiceField):
    """A ChoiceField whose choices are a sqlalchemy model list relationship."""

    widget = djangofields.SelectMultiple
    hidden_widget = djangofields.MultipleHiddenInput
    default_error_messages = {
        "list": gettext_lazy("Enter a list of values."),
        "invalid_choice": gettext_lazy("Select a valid choice. %(value)s is not one of the available choices."),
    }

    def __init__(self, *args, **kwargs):
        kwargs["empty_label"] = None
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        return list(self._check_values(value)) if value else []

    def _check_values(self, value):
        return [self.get_object(pk) for pk in value]

    def prepare_value(self, value):
        try:
            return [self.prepare_instance_value(v) for v in value if not isinstance(v, str)]

        except Exception:
            return super().prepare_value(value)



import unittest

import attr
from django_sorcery.db import SQLAlchemy, databases, middleware

from ..base import mock


class DummyMiddeware(middleware.BaseMiddleware):
    rollback_called = False
    flush_called = False
    commit_called = False
    remove_called = False

    def process_request(self, request):
        super().process_request(request)
        if hasattr(self, "dummy"):
            return self.dummy

    def rollback(self, request, response):
        self.rollback_called = True

    def flush(self, request, response):
        self.flush_called = True
        if hasattr(self, "flush_error"):
            raise self.flush_error()

    def commit(self, request, response):
        self.commit_called = True

    def remove(self, request, response):
        self.remove_called = True


def get_response(request):
    return request


@attr.s
class Request:
    method = attr.ib(default="GET")
    status_code = attr.ib(default=200)


class TestBaseMiddleware(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.middleware = DummyMiddeware(get_response=get_response)

    def test_process_request(self):
        self.middleware.dummy = "abcd"
        response = self.middleware(Request())
        self.assertIsNotNone(response)
        self.assertFalse(self.middleware.rollback_called)
        self.assertFalse(self.middleware.flush_called)
        self.assertFalse(self.middleware.commit_called)
        self.assertTrue(self.middleware.remove_called)

    def test_success(self):
        response = self.middleware(Request())
        self.assertIsNotNone(response)
        self.assertFalse(self.middleware.rollback_called)
        self.assertTrue(self.middleware.flush_called)
        self.assertTrue(self.middleware.commit_called)
        self.assertTrue(self.middleware.remove_called)

    def test_redirect(self):
        response = self.middleware(Request(status_code=300))
        self.assertIsNotNone(response)
        self.assertFalse(self.middleware.rollback_called)
        self.assertTrue(self.middleware.flush_called)
        self.assertTrue(self.middleware.commit_called)
        self.assertTrue(self.middleware.remove_called)

    def test_bad_request(self):
        response = self.middleware(Request(status_code=400))
        self.assertIsNotNone(response)
        self.assertTrue(self.middleware.rollback_called)
        self.assertFalse(self.middleware.flush_called)
        self.assertFalse(self.middleware.commit_called)
        self.assertTrue(self.middleware.remove_called)

    def test_server_error(self):
        response = self.middleware(Request(status_code=400))
        self.assertIsNotNone(response)
        self.assertTrue(self.middleware.rollback_called)
        self.assertFalse(self.middleware.flush_called)
        self.assertFalse(self.middleware.commit_called)
        self.assertTrue(self.middleware.remove_called)

    def test_flush_error(self):
        self.middleware.flush_error = RuntimeError
        with self.assertRaises(RuntimeError):
            self.middleware(Request(status_code=200))
        self.assertTrue(self.middleware.rollback_called)
        self.assertTrue(self.middleware.flush_called)
        self.assertFalse(self.middleware.commit_called)
        self.assertTrue(self.middleware.remove_called)

    def test_other_methods(self):
        response = self.middleware(Request(method="HEAD"))
        self.assertIsNotNone(response)
        self.assertTrue(self.middleware.rollback_called)
        self.assertFalse(self.middleware.flush_called)
        self.assertFalse(self.middleware.commit_called)
        self.assertTrue(self.middleware.remove_called)


class TestSQLAlchemyDBMiddleware(unittest.TestCase):
    def setUp(self):
        super().setUp()

        self.db = mock.MagicMock()
        self.middleware = type("DummySQLAlchemyMiddleware", (middleware.SQLAlchemyDBMiddleware,), {"db": self.db})(
            get_response
        )

    def test_commit(self):
        self.middleware.commit(None, None)
        self.db.commit.assert_called_once_with()

    def test_rollback(self):
        self.middleware.rollback(None, None)
        self.db.rollback.assert_called_once_with()

    def test_flush(self):
        self.middleware.flush(None, None)
        self.db.flush.assert_called_once_with()

    def test_remove(self):
        self.middleware.remove(None, None)
        self.db.remove.assert_called_once_with()


class TestSQLAlchemyMiddleware(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.middleware = middleware.SQLAlchemyMiddleware(get_response)
        self.orig_dbs = databases.copy()
        databases.clear()
        databases.update(one=mock.MagicMock(spec=SQLAlchemy), two=mock.MagicMock(spec=SQLAlchemy))

    def tearDown(self):
        super().setUp()
        databases.clear()
        databases.update(self.orig_dbs)

    def test_commit(self):
        self.middleware.commit(None, None)
        for db in databases.values():
            db.commit.assert_called_once_with()

    def test_rollback(self):
        self.middleware.rollback(None, None)
        for db in databases.values():
            db.rollback.assert_called_once_with()

    def test_flush(self):
        self.middleware.flush(None, None)
        for db in databases.values():
            db.flush.assert_called_once_with()

    def test_remove(self):
        self.middleware.remove(None, None)
        for db in databases.values():
            db.remove.assert_called_once_with()



"""Metadata for composite sqlalchemy properties."""
from collections import OrderedDict

import sqlalchemy as sa
from django.core.exceptions import ValidationError

from ...exceptions import NestedValidationError
from ...utils import get_args
from ...validators import ValidationRunner
from .base import model_info_meta
from .column import column_info


class composite_info(metaclass=model_info_meta):
    """A helper class that makes sqlalchemy composite model inspection
    easier."""

    __slots__ = ("prop", "properties", "parent", "_field_names")

    def __init__(self, composite, parent=None):
        self._field_names = set()
        self.prop = composite.prop
        self.parent = parent

        attrs = [k for k, v in sorted(vars(self.prop.composite_class).items()) if isinstance(v, sa.Column)]
        if not attrs:
            attrs = get_args(self.prop.composite_class.__init__)

        self.properties = OrderedDict()
        for attr, prop, col in zip(attrs, self.prop.props, self.prop.columns):
            self.properties[attr] = column_info(col, prop, self, name=attr)

    @property
    def field_names(self):
        """Returns field names used in composite."""
        if not self._field_names:
            self._field_names.update(self.properties.keys())

            self._field_names = [attr for attr in self._field_names if not attr.startswith("_")]

        return self._field_names

    @property
    def name(self):
        """Returns composite field name."""
        return self.prop.key

    @property
    def attribute(self):
        """Returns composite field instrumented attribute for generating query
        expressions."""
        return getattr(self.parent_model, self.name)

    @property
    def parent_model(self):
        """Returns the model class that the attribute belongs to."""
        return self.prop.parent.class_

    @property
    def model_class(self):
        """Returns the composite class."""
        return self.prop.composite_class

    def __repr__(self):
        reprs = [
            "<composite_info({!s}, {!s}.{!s})>".format(self.model_class.__name__, self.parent_model.__name__, self.name)
        ]
        reprs.extend("    " + repr(i) for _, i in sorted(self.properties.items()))
        return "\n".join(reprs)

    def clean_fields(self, instance, exclude=None):
        """Clean all fields and raise a ValidationError containing a dict of
        all validation errors if any occur."""
        errors = {}
        exclude = exclude or []
        for name, f in self.properties.items():
            raw_value = getattr(instance, name, None)
            is_blank = not bool(raw_value)
            is_nullable = f.null
            is_defaulted = f.column.default or f.column.server_default
            is_required = f.required

            is_skippable = is_blank and (is_nullable or is_defaulted or not is_required)

            if name in exclude or is_skippable:
                continue
            try:
                setattr(instance, name, f.clean(raw_value, instance))
            except ValidationError as e:
                errors[name] = e.error_list
        if errors:
            raise NestedValidationError(errors)

    def run_validators(self, instance):
        """Run composite field's validators and raise ValidationError if
        necessary."""
        runner = ValidationRunner(validators=getattr(instance, "validators", []))
        runner.is_valid(instance, raise_exception=True)

    def full_clean(self, instance, exclude=None):
        """Call clean_fields(), clean(), and run_validators() on the composite
        model.

        Raise a ValidationError for any errors that occur.
        """
        runner = ValidationRunner(
            name=self.name,
            validators=[
                lambda x: self.clean_fields(x, exclude),
                self.run_validators,
                lambda x: getattr(x, "clean", bool)(),
            ],
        )
        runner.is_valid(instance, raise_exception=True)



from django_sorcery import routers

from .testapp import views


try:
    from django.conf.urls import url as re_path
    from django.conf.urls import include
except ImportError:  # pragma: no cover
    from django.urls import re_path, include


router = routers.SimpleRouter()
router.register("owners", views.OwnerViewSet)

urlpatterns = [
    # list views
    # r"^(?P<id>[0-9]+)/$"
    re_path(r"^list/owners/$", views.OwnerListView.as_view(), name="owners_list"),
    re_path(r"^list/owners-tmpl/$", views.OwnerListViewWithTemplate.as_view(), name="owners_list_tmpl"),
    re_path(r"^list/owners-order/$", views.OwnerListViewWithOrdering.as_view(), name="owners_list_order"),
    re_path(
        r"^list/owners-context-name/$", views.OwnerListViewNoContextName.as_view(), name="owners_list_context_name"
    ),
    re_path(
        r"^list/owners-no-empty-paginate/$",
        views.OwnerListViewNoEmptyPaginateBy.as_view(),
        name="owners_list_no_empty_paginate",
    ),
    re_path(
        r"^list/dummy-no-empty-paginate/$",
        views.DummyListViewNoEmptyPaginateBy.as_view(),
        name="dummy_list_no_empty_paginate",
    ),
    # detail views
    re_path(r"^detail/owners/(?P<id>[0-9]+)/$", views.OwnerDetailView.as_view(), name="owner_detail"),
    re_path(
        r"^detail/owners-tmpl/(?P<id>[0-9]+)/$", views.OwnerDetailViewWithTemplate.as_view(), name="owner_detail_tmpl"
    ),
    re_path(r"^detail/owners-slug/(?P<slug>\w+)/", views.OwnerDetailViewWithSlug.as_view(), name="owner_detail_slug"),
    re_path(
        r"^detail/owners-context-name/(?P<id>[0-9]+)/$",
        views.OwnerDetailViewContextName.as_view(),
        name="owner_detail_context_name",
    ),
    # edit views
    re_path(r"^edit/owners/create/$", views.OwnerCreateView.as_view(), name="owner_create"),
    re_path(r"^edit/owners/(?P<id>[0-9]+)/$", views.OwnerUpdateView.as_view(), name="owner_update"),
    re_path(
        r"^edit/owners/create-field-form/$",
        views.OwnerCreateViewWithFieldsAndForm.as_view(),
        name="owner_create_field_form",
    ),
    re_path(r"^edit/owners/create-form/$", views.OwnerCreateViewWithForm.as_view(), name="owner_create_form"),
    re_path(r"^edit/vehicles/create/$", views.VehicleCreateView.as_view(), name="vehicle_create"),
    # delete views
    re_path(r"^delete/owners/(?P<id>[0-9]+)/$", views.OwnerDeleteView.as_view(), name="owner_delete"),
    re_path(r"^viewsets/", include(router.urls)),
]



"""sqlalchemy relationship related things."""
from contextlib import suppress
from itertools import chain

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr

from ..utils import setdefaultattr
from .signals import declare_first


class RelationsMixin:
    """Mixin that provides django like shortcuts for relationships."""

    def _one_relation(self, remote_cls, direction, backref_uselist, kwargs):
        @declared_attr
        def o2m(cls):
            rels = setdefaultattr(cls, "_relationships", set())
            kwargs.setdefault("info", {}).update(self._get_kwargs_for_relation(kwargs))
            backref = kwargs.get("backref")
            if backref:
                backref_kwargs = None
                if isinstance(backref, tuple):
                    with suppress(Exception):
                        backref, backref_kwargs = backref

                backref_kwargs = backref_kwargs or {}
                backref_kwargs["uselist"] = backref_uselist

                kwargs["backref"] = self.backref(backref, **backref_kwargs)

            rel = self.relationship(remote_cls, **kwargs)
            rel.direction = direction
            rels.add(rel)
            return rel

        return o2m

    def OneToOne(self, remote_cls, **kwargs):
        """Use an event to build one-to-many relationship on a model and auto
        generates foreign key relationship from the remote table::

        class ModelOne(db.Model):
            pk = db.Column(.., primary_key=True)
            m2 = db.OneToOne("ModelTwo", ...)

        class ModelTwo(db.Model):
            pk = db.Column(.., primary_key=True)
            ...

        will create ModelTwo.m1_pk automatically for the relationship
        """
        kwargs["uselist"] = False
        return self._one_relation(
            remote_cls, direction=sa.orm.interfaces.MANYTOONE, backref_uselist=False, kwargs=kwargs
        )

    def OneToMany(self, remote_cls, **kwargs):
        """Use an event to build one-to-many relationship on a model and auto
        generates foreign key relationship from the remote table::

        class ModelOne(db.Model):
            pk = db.Column(.., primary_key=True)
            m2 = db.OneToMany("ModelTwo", ...)

        class ModelTwo(db.Model):
            pk = db.Column(.., primary_key=True)
            ...

        will create ModelTwo.m1_pk automatically for the relationship
        """
        kwargs["uselist"] = True
        return self._one_relation(
            remote_cls, direction=sa.orm.interfaces.ONETOMANY, backref_uselist=False, kwargs=kwargs
        )

    def ManyToOne(self, remote_cls, **kwargs):
        """Use an event to build many-to-one relationship on a model and auto
        generates foreign key relationship on the remote table::

            class ModelOne(db.Model):
                pk = db.Column(.., primary_key=True)
                m2 = db.ManyToOne("ModelTwo", ...)

            class ModelTwo(db.Model):
                pk = db.Column(.., primary_key=True)
                ...

        will create ModelOne.m2_pk automatically for the relationship
        """
        kwargs["uselist"] = False
        return self._one_relation(
            remote_cls, direction=sa.orm.interfaces.MANYTOONE, backref_uselist=True, kwargs=kwargs
        )

    def ManyToMany(self, remote_cls, table_name=None, **kwargs):
        """Use an event to build many-to-many relationship on a model and auto
        generates an association table or if a model is provided as secondary
        argument::

            class ModelOne(db.Model):
                pk = db.Column(.., primary_key=True)
                m2s = db.ManyToMany("ModelTwo", backref="m1s", table_name='m1m2s', ...)

            class ModelTwo(db.Model):
                pk = db.Column(.., primary_key=True)
                ...

        or with back_populates::

            class ModelOne(db.Model):
                pk = db.Column(.., primary_key=True)
                m2s = db.ManyToMany("ModelTwo", back_populates="m1s", table_name='m1m2s', ...)

            class ModelTwo(db.Model):
                pk = db.Column(.., primary_key=True)
                m1s = db.ManyToMany("ModelOne", back_populates="m2s", table_name='m1m2s', ...)

        will create ModelOne.m2s and ModelTwo.m1s relationship thru a provided secondary argument.
        If no secondary argument is provided, table_name is required as it will be used for the autogenerated
        association table.

        In the case of back_populates you have to provide the same table_name argument on both many-to-many
        declarations
        """

        @declared_attr
        def m2m(cls):
            """many to many relationship attribute for declarative."""
            if "secondary" not in kwargs and table_name is None:
                raise sa.exc.ArgumentError(
                    "You need to provide secondary or table_name for the relation for the association table "
                    "that will be generated"
                )

            rels = setdefaultattr(cls, "_relationships", set())
            info = kwargs.setdefault("info", {})
            info.update(self._get_kwargs_for_relation(kwargs))
            info.update(self._get_kwargs_for_relation(kwargs, "table_"))
            if table_name:
                info["table_name"] = table_name

            kwargs["uselist"] = True

            backref = kwargs.get("backref")
            backref_kwargs = None
            if backref:
                if isinstance(backref, tuple):
                    with suppress(Exception):
                        backref, backref_kwargs = backref

                backref_kwargs = backref_kwargs or {}

                backref_kwargs["uselist"] = True
                kwargs["backref"] = self.backref(backref, **backref_kwargs)

            rel = self.relationship(remote_cls, **kwargs)
            rel.direction = sa.orm.interfaces.MANYTOMANY
            rels.add(rel)
            return rel

        return m2m

    def _get_kwargs_for_relation(self, kwargs, prefix="fk_"):
        return {key: kwargs.pop(key) for key in list(kwargs.keys()) if key.startswith(prefix)}


def _add_foreign_keys(cls, parent_cls, relation):
    """Generate fk columns and constraint to the remote class from a
    relationship."""
    fk_kwargs = {key[3:]: val for key, val in relation.info.items() if key.startswith("fk_")}
    fk_prefix = fk_kwargs.pop("prefix", "_")
    fk_nullable = fk_kwargs.pop("nullable", True)
    fk_key = fk_kwargs.pop("key", None)

    if not fk_key:
        if relation.direction == sa.orm.interfaces.MANYTOONE:
            fk_key = relation.key.lower()
        elif relation.backref:
            backref, _ = relation.backref
            fk_key = backref.lower()
        else:
            fk_key = parent_cls.__name__.lower()

    cols = {}
    cols_created = False
    for pk_column in parent_cls.__table__.primary_key:
        pk_attr = parent_cls.__mapper__.get_property_by_column(pk_column)
        col_name = "_".join(filter(None, [fk_key, pk_column.name]))
        attr = "{}{}".format(fk_prefix, "_".join(filter(None, [fk_key, pk_attr.key])))

        if col_name not in cls.__table__.columns and not hasattr(cls, attr):
            fk_column = sa.Column(col_name, pk_column.type, nullable=fk_nullable)
            setattr(cls, attr, fk_column)
            cols_created = True
        else:
            fk_column = cls.__table__.columns[col_name]

        cols[pk_column] = fk_column

    relation._user_defined_foreign_keys = cols.values()

    if cols_created:
        # pk and fk ordering must match for foreign key constraint
        pks, fks = [], []
        for pk, value in cols.items():
            pks.append(pk)
            fks.append(value)

        constraint = sa.ForeignKeyConstraint(fks, pks, **fk_kwargs)
        cls.__table__.append_constraint(constraint)


def _add_association_table(cls, child_cls, relation):
    """Generate association table and fk constraints to satisfy a many-to-many
    relation."""
    if relation.secondary is not None:
        return

    table_name = relation.info.get("table_name")
    relation.secondary = cls.metadata.tables.get(table_name)
    if relation.secondary is not None:
        return

    fk_kwargs = {key[3:]: val for key, val in relation.info.items() if key.startswith("fk_")}
    table_kwargs = {key[6:]: val for key, val in relation.info.items() if key.startswith("table_")}
    table_kwargs.pop("name", None)

    column_map = {}
    for pk_column in chain(cls.__mapper__.primary_key, child_cls.__table__.primary_key):
        col_name = "_".join(filter(None, [pk_column.table.name.lower(), pk_column.name]))
        col = sa.Column(col_name, pk_column.type, primary_key=True)
        column_map.setdefault(pk_column.table, []).append(col)

    table_args = list(chain(*column_map.values()))

    for table, columns in column_map.items():
        table_args.append(sa.ForeignKeyConstraint(columns, table.primary_key, **fk_kwargs))

    relation.secondary = sa.Table(table_name, cls.metadata, *table_args, schema=cls.__table__.schema, **table_kwargs)


@declare_first.connect
def declare_first_relationships_handler(cls):
    """Declare first signal handler which connects relationships on the class.

    Can be called multiple times so once relationships are set, they are
    removed from model
    """
    rels = getattr(cls, "_relationships", set())

    for relation in rels:
        if relation.direction == sa.orm.interfaces.ONETOMANY:
            _add_foreign_keys(relation.mapper.class_, cls, relation)
        elif relation.direction == sa.orm.interfaces.MANYTOONE:
            _add_foreign_keys(cls, relation.mapper.class_, relation)
        elif relation.direction == sa.orm.interfaces.MANYTOMANY:
            _add_association_table(cls, relation.mapper.class_, relation)

    rels.clear()



from django.core.exceptions import ImproperlyConfigured
from django_sorcery import viewsets
from django_sorcery.routers import BaseRouter, Route, SimpleRouter, action

from .base import TestCase
from .testapp.models import Owner
from .testapp.views import OwnerViewSet


class TestBaseRouter(TestCase):
    def test_router(self):
        router = BaseRouter()

        self.assertEqual(router.registry, [])

        router.register("prefix", OwnerViewSet, base_name="base_name")

        self.assertEqual(router.registry, [("prefix", OwnerViewSet, "base_name")])

        with self.assertRaises(NotImplementedError):
            router.register("prefix", OwnerViewSet)

        with self.assertRaises(NotImplementedError):
            router.get_default_base_name(OwnerViewSet)

        with self.assertRaises(NotImplementedError):
            router.get_urls()

        with self.assertRaises(NotImplementedError):
            router.urls

        router._urls = "test"
        self.assertEqual(router.urls, "test")


class TestSimpleRouter(TestCase):
    def get_regex(self, pattern):
        try:
            return pattern.pattern._regex

        except AttributeError:
            return pattern._regex

    def test_router(self):
        router = SimpleRouter()

        router.register("prefix", OwnerViewSet)

        self.assertEqual(router.registry, [("prefix", OwnerViewSet, "owner")])

        self.assertEqual(
            router.get_routes(OwnerViewSet),
            [
                Route(
                    url="^{prefix}{trailing_slash}$",
                    mapping={"get": "list", "post": "create"},
                    name="{basename}-list",
                    detail=False,
                    initkwargs={"suffix": "List"},
                ),
                Route(
                    url="^{prefix}/new{trailing_slash}$",
                    mapping={"get": "new"},
                    name="{basename}-new",
                    detail=False,
                    initkwargs={"suffix": "New"},
                ),
                Route(
                    url="^{prefix}/{lookup}{trailing_slash}$",
                    mapping={
                        "get": "retrieve",
                        "post": "update",
                        "put": "update",
                        "patch": "update",
                        "delete": "destroy",
                    },
                    name="{basename}-detail",
                    detail=True,
                    initkwargs={"suffix": "Instance"},
                ),
                Route(
                    url="^{prefix}/{lookup}/edit{trailing_slash}$",
                    mapping={"get": "edit"},
                    name="{basename}-edit",
                    detail=True,
                    initkwargs={"suffix": "Instance"},
                ),
                Route(
                    url="^{prefix}/{lookup}/delete{trailing_slash}$",
                    mapping={"get": "confirm_destroy", "post": "destroy"},
                    name="{basename}-destroy",
                    detail=True,
                    initkwargs={"suffix": "Instance"},
                ),
                Route(
                    url="^{prefix}/{lookup}/custom{trailing_slash}$",
                    mapping={"get": "custom"},
                    name="{basename}-custom",
                    detail=True,
                    initkwargs={},
                ),
            ],
        )

        urls = {url.name: url for url in router.urls}

        self.assertEqual(self.get_regex(urls["owner-custom"]), "^prefix/(?P<id>[^/.]+)/custom/$")
        self.assertEqual(self.get_regex(urls["owner-destroy"]), "^prefix/(?P<id>[^/.]+)/delete/$")
        self.assertEqual(self.get_regex(urls["owner-detail"]), "^prefix/(?P<id>[^/.]+)/$")
        self.assertEqual(self.get_regex(urls["owner-edit"]), "^prefix/(?P<id>[^/.]+)/edit/$")
        self.assertEqual(self.get_regex(urls["owner-list"]), "^prefix/$")
        self.assertEqual(self.get_regex(urls["owner-new"]), "^prefix/new/$")

    def test_action_on_existing_action(self):
        class OwnerViewSet(viewsets.ListModelMixin, viewsets.GenericViewSet):
            model = Owner

            @action(detail=False)
            def list(self, request, *args, **kwargs):
                pass

        router = SimpleRouter()

        with self.assertRaises(ImproperlyConfigured):
            router.get_routes(OwnerViewSet)

    def test_custom_lookup_regex(self):
        class OwnerViewSet(viewsets.RetrieveModelMixin, viewsets.GenericViewSet):
            model = Owner
            lookup_url_regex = "lookup_regex"

        router = SimpleRouter()

        self.assertEqual(router.get_lookup_regex(OwnerViewSet), "lookup_regex")

    def test_lookup_regex_no_model(self):
        class OwnerViewSet(viewsets.GenericViewSet):
            pass

        router = SimpleRouter()
        self.assertEqual(router.get_lookup_regex(OwnerViewSet), "(?P<pk>[^/.]+)")

    def test_path_with_mustaches(self):
        class OwnerViewSet(viewsets.GenericViewSet):
            @action(detail=False, url_path="{dummy}")
            def dummy(self, request, *args, **kwargs):
                pass

        router = SimpleRouter()
        routes = router.get_routes(OwnerViewSet)
        dummy_route = next(iter(filter(lambda r: "dummy" in r.url, routes)), None)
        self.assertIsNotNone(dummy_route)
        self.assertEqual(dummy_route.url, "^{prefix}/{{dummy}}{trailing_slash}$")

    def test_get_url_with_empty_prefix(self):
        class OwnerViewSet(viewsets.ListModelMixin, viewsets.GenericViewSet):
            model = Owner

        router = SimpleRouter()
        router.register("", OwnerViewSet, "")
        urlpattern = router.get_urls()[0]
        self.assertEqual(self.get_regex(urlpattern), "^$")



#!/usr/bin/env python
# fmt:off
import os

from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))


about = {}
with open(os.path.join(here, "django_sorcery", "__version__.py")) as f:
    exec(f.read(), about)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), "rb").read().decode("utf-8")


setup(
    author=about["__author__"],
    author_email=about["__author_email__"],
    description=about["__description__"],
    include_package_data=True,
    install_requires=["alembic", "blinker", "inflect", "django", "python-dateutil", "sqlalchemy<2"],
    license="MIT",
    long_description=read("README.rst"),
    name="django-sorcery",
    packages=find_packages(exclude=["tests", "tests.*"]),
    entry_points={"pytest11": ["django-sorcery = django_sorcery.pytest_plugin"]},
    url="https://github.com/shosca/django-sorcery",
    version=about["__version__"],
    keywords="sqlalchemy django framework forms",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Documentation": "https://django-sorcery.readthedocs.io",
        "Source": "https://github.com/shosca/django-sorcery",
    },
)



from django_sorcery.db import databases


db = databases.get("default")


class Question(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    question_text = db.Column(db.String(length=200))
    pub_date = db.Column(db.DateTime())


class Choice(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    choice_text = db.Column(db.String(length=200))
    votes = db.Column(db.Integer(), default=0)

    question = db.ManyToOne(Question, backref=db.backref("choices", cascade="all, delete-orphan"))



from io import StringIO

from django.core.management.base import BaseCommand
from django_sorcery.management import NamespacedCommand

from ..base import TestCase


class SubCommand(BaseCommand):
    help = "subcommand"

    def add_arguments(self, parser):
        parser.add_argument("-d", action="append", dest="dummy", type=str, help="subcommand help message")

    def handle(self, *args, **kwargs):
        return "subcommand triggered"


class DummyCommand(NamespacedCommand):
    sub1 = SubCommand

    class Meta:
        namespace = "dummy"


class TestNamespacedCommand(TestCase):
    def setUp(self):
        super().setUp()
        self.stdout = StringIO()
        self.stderr = StringIO()
        self.cmd = DummyCommand(stdout=self.stdout, stderr=self.stderr, no_color=True)

    def test_help_message(self):
        self.cmd.run_from_argv(["./manage.py", "dummy", "-h"])

        self.stdout.seek(0)
        out = "\n".join([line.strip() for line in self.stdout.readlines()])
        self.assertTrue(out.startswith("usage: manage.py dummy [-h] sub1"))

    def test_trigger_subcommand(self):
        self.cmd.run_from_argv(["./manage.py", "dummy", "sub1"])

        self.stdout.seek(0)
        self.assertEqual("".join([line.strip() for line in self.stdout.readlines()]), "subcommand triggered")



from django.http import HttpResponseRedirect
from django.urls import reverse, reverse_lazy
from django_sorcery.formsets import inlineformset_factory
from django_sorcery.routers import action
from django_sorcery.viewsets import ModelViewSet

from .models import Choice, Question, db


ChoiceFormSet = inlineformset_factory(relation=Question.choices, fields=(Choice.choice_text.key,), session=db)


class PollsViewSet(ModelViewSet):
    model = Question
    fields = (Question.question_text.key, Question.pub_date.key)
    destroy_success_url = reverse_lazy("polls:question-list")

    def get_success_url(self):
        return reverse("polls:question-detail", kwargs={"pk": self.object.pk})

    def get_form_context_data(self, **kwargs):
        kwargs["choice_set"] = self.get_choice_formset()
        return super().get_form_context_data(**kwargs)

    def get_choice_formset(self, instance=None):
        if not hasattr(self, "_choice_formset"):
            instance = instance or self.object
            self._choice_formset = ChoiceFormSet(
                instance=instance, data=self.request.POST if self.request.POST else None
            )

        return self._choice_formset

    def process_form(self, form):
        if form.is_valid() and self.get_choice_formset(instance=form.instance).is_valid():
            return self.form_valid(form)

        return form.invalid(self, form)

    def form_valid(self, form):
        self.object = form.save()
        self.object.choices = self.get_choice_formset().save()
        db.flush()
        return HttpResponseRedirect(self.get_success_url())

    @action(detail=True)
    def results(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    @action(detail=True, methods=["POST"])
    def vote(self, request, *args, **kwargs):
        self.object = self.get_object()

        selected_choice = Choice.query.filter(
            Choice.question == self.object, Choice.pk == request.POST.get("choice")
        ).one_or_none()

        if not selected_choice:
            context = self.get_detail_context_data(object=self.object)
            context["error_message"] = "You didn't select a choice."
            self.action = "retrieve"
            return self.render_to_response(context)

        selected_choice.votes += 1
        db.flush()
        return HttpResponseRedirect(reverse("polls:question-results", args=(self.object.pk,)))



"""Support for reusable sqlalchemy composite fields."""
from collections import OrderedDict

import sqlalchemy as sa
from sqlalchemy.orm import CompositeProperty
from sqlalchemy.util.langhelpers import classproperty


class CompositeField(CompositeProperty):
    """Composite field which understands composite objects with builtin
    columns.

    See :py:class:`.BaseComposite` for examples.
    """

    def __init__(self, class_, **kwargs):
        self.random_prefix = class_.__name__ + str(id(self))
        self.prefix = kwargs.pop("prefix", None) or self.random_prefix

        columns = OrderedDict()
        for k, c in class_._columns.items():
            columns[k] = c = c.copy()
            c.name = self.prefix + "_" + (c.name or k)
            c.key = "_" + c.name

            # if column type has name to add constraints we need to account for that
            if hasattr(c.type, "name") and c.type.name:
                c.type.name = self.prefix + "_" + c.type.name

        super().__init__(class_, *list(columns.values()), **kwargs)

    def instrument_class(self, mapper):
        if self.prefix == self.random_prefix:
            for c in self.attrs:
                c.name = self.key + c.name.replace(self.random_prefix, "")
                c.key = "_" + self.key + c.key.replace(self.random_prefix, "")[1:]

                if hasattr(c.type, "name") and c.type.name:
                    c.type.name = c.type.name.replace(self.random_prefix, "")[1:]

        return super().instrument_class(mapper)


class BaseComposite:
    """Base class for creating composite classes which
    :py:class:`.CompositeField` will understand.

    For example::

        class MyComposite(object):
            foo = db.Column(db.Integer())
            bar = db.Column(db.Integer())

        class MyModel(db.Model):
            test = db.CompositeField(MyComposite)
            # both test_foo and test_bar columns will be added to model
            # their instrumented properties will be _test_foo and _test_bar
    """

    def __init__(self, *args, **kwargs):
        for k in self._columns:
            setattr(self, k, None)
        for k, v in zip(self._columns, args):
            setattr(self, k, v)
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classproperty
    def _columns(cls):
        return OrderedDict([(k, v) for k, v in sorted(vars(cls).items()) if isinstance(v, sa.Column)])

    def __composite_values__(self):
        return tuple(getattr(self, i) for i in self._columns)

    def as_dict(self):
        """Serializer composite to a dictionary."""
        return OrderedDict([(k, getattr(self, k)) for k in self._columns])

    def __repr__(self):
        return "".join(
            [
                self.__class__.__name__,
                "(",
                ", ".join("{}={!r}".format(k, v) for k, v in self.as_dict().items() if v),
                ")",
            ]
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__composite_values__() == other.__composite_values__()

    def __bool__(self):
        return any(vars(self).values())

    __nonzero__ = __bool__



import sqlalchemy as sa
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django_sorcery.views.base import SQLAlchemyMixin

from ..testapp.models import ClassicModel, Owner, db


class TestBaseView(TestCase):
    def test_get_model(self):
        class DummyView(SQLAlchemyMixin):
            model = None

        with self.assertRaises(ImproperlyConfigured):
            DummyView.get_model()

    def test_get_queryset(self):
        class DummyView(SQLAlchemyMixin):
            model = Owner
            queryset = Owner.query

        view = DummyView()

        query = view.get_queryset()

        self.assertIsInstance(query, sa.orm.Query)
        self.assertEqual(query._only_full_mapper_zero("get").class_, Owner)

    def test_get_queryset_from_model_only(self):
        class DummyView(SQLAlchemyMixin):
            model = Owner

        view = DummyView()

        query = view.get_queryset()

        self.assertIsInstance(query, sa.orm.Query)
        self.assertEqual(query._only_full_mapper_zero("get").class_, Owner)

    def test_get_queryset_from_session(self):
        class DummyView(SQLAlchemyMixin):
            model = Owner
            session = db

        view = DummyView()

        query = view.get_queryset()

        self.assertIsInstance(query, sa.orm.Query)
        self.assertEqual(query._only_full_mapper_zero("get").class_, Owner)

    def test_get_queryset_with_options(self):
        class DummyViewWithOptions(SQLAlchemyMixin):
            model = Owner
            session = db
            query_options = [sa.orm.noload("*")]

        view = DummyViewWithOptions()

        query = view.get_queryset()

        self.assertIsInstance(query, sa.orm.Query)
        self.assertEqual(query._only_full_mapper_zero("get").class_, Owner)
        self.assertEqual(len(query._with_options), 1)

    def test_get_queryset_fail(self):
        class DummyObject:
            pass

        class DummyViewFail(SQLAlchemyMixin):
            model = DummyObject

        view = DummyViewFail()

        with self.assertRaises(ImproperlyConfigured) as ctx:
            view.get_queryset()

        self.assertEqual(
            str(ctx.exception),
            "DummyViewFail is missing a QuerySet. Define DummyViewFail.model and DummyViewFail.session, "
            "DummyViewFail.queryset, or override DummyViewFail.get_queryset().",
        )

    def test_get_queryset_classical_mapping(self):
        class ClassicModelView(SQLAlchemyMixin):
            model = ClassicModel
            session = db

        view = ClassicModelView()

        query = view.get_queryset()

        self.assertIsInstance(query, sa.orm.Query)
        self.assertEqual(query._only_full_mapper_zero("get").class_, ClassicModel)

    def test_get_session(self):
        class OwnerView(SQLAlchemyMixin):
            model = Owner
            queryset = Owner.query

        view = OwnerView()

        view.get_session()

        self.assertEqual(view.session, db)



import enum

import sqlalchemy as sa
from django import forms as djangoforms
from django.core import validators as django_validators
from django.forms import fields as djangofields
from django_sorcery import fields as formfields
from django_sorcery.db import fields, meta

from ..base import TestCase, mock


class DummyEnum(enum.Enum):
    one = 1
    two = 2


class TestFields(TestCase):
    def test_can_init_like_sqlalchemy(self):
        col_type = sa.String()
        column = fields.Field("test", col_type)

        self.assertEqual(column.name, "test")
        self.assertEqual(column.type, col_type)

    def test_boolean_field(self):
        column = fields.BooleanField()
        self.assertIsInstance(column.type, sa.Boolean)
        self.assertFalse(column.nullable)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.BooleanField)

        column = fields.BooleanField(constraint_name="test")
        self.assertEqual(column.type.name, "test")

    def test_char_field(self):
        with self.assertRaises(TypeError):
            column = fields.CharField()

        column = fields.CharField(max_length=120)
        self.assertEqual(column.type.length, 120)
        self.assertIsInstance(column.info["validators"][-1], django_validators.MaxLengthValidator)
        self.assertEqual(column.info["validators"][-1].limit_value, 120)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.CharField)

    def test_date_field(self):
        column = fields.DateField()
        self.assertIsInstance(column.type, sa.Date)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.DateField)

    def test_datetime_field(self):
        column = fields.DateTimeField()
        self.assertIsInstance(column.type, sa.DateTime)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.DateTimeField)

    def test_duration_field(self):
        column = fields.DurationField()
        self.assertIsInstance(column.type, sa.Interval)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.DurationField)

    def test_decimal_field(self):
        column = fields.DecimalField()
        self.assertIsInstance(column.type, sa.Numeric)
        self.assertTrue(column.type.asdecimal)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.DecimalField)

    def test_email_field(self):
        column = fields.EmailField(length=100)
        self.assertIsInstance(column.type, sa.String)
        self.assertEqual(
            {
                "form_class": djangofields.EmailField,
                "required": False,
                "validators": [django_validators.validate_email, mock.ANY],
            },
            column.info,
        )
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.EmailField)

    def test_enum_field_bad_config(self):
        with self.assertRaises(TypeError):
            fields.EnumField()

        with self.assertRaises(TypeError):
            fields.EnumField(DummyEnum)

    def test_enum_field_plain_choices(self):
        column = fields.EnumField(choices=["a", "b"])
        self.assertIsInstance(column.type, sa.Enum)
        self.assertIsNone(column.type.enum_class)
        self.assertEqual(column.type.enums, ["a", "b"])
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.TypedChoiceField)
        self.assertEqual(form_field.choices, [("a", "a"), ("b", "b")])

    def test_enum_field_enum_choices(self):
        column = fields.EnumField(choices=DummyEnum)
        self.assertIsInstance(column.type, sa.Enum)
        self.assertEqual(column.type.enum_class, DummyEnum)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, formfields.EnumField)

    def test_enum_field_enum_choices_with_constraint_name(self):
        column = fields.EnumField(choices=DummyEnum, constraint_name="dummy")
        self.assertIsInstance(column.type, sa.Enum)
        self.assertEqual(column.type.enum_class, DummyEnum)
        self.assertEqual(column.type.name, "dummy")
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, formfields.EnumField)

    def test_enum_field_enum_choices_with_constraint_name_and_no_native_type(self):
        column = fields.EnumField(choices=DummyEnum, constraint_name="dummy", native_enum=False)
        self.assertIsInstance(column.type, sa.Enum)
        self.assertEqual(column.type.enum_class, DummyEnum)
        self.assertEqual(column.type.name, "dummy")
        self.assertFalse(column.type.native_enum)
        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, formfields.EnumField)

    def test_float_field(self):
        column = fields.FloatField()
        self.assertIsInstance(column.type, sa.Float)
        self.assertIsNone(column.type.precision)
        self.assertFalse(column.type.asdecimal)

        column = fields.FloatField(precision=10)
        self.assertEqual(column.type.precision, 10)

        column = fields.FloatField(max_digits=10)
        self.assertEqual(column.type.precision, 10)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.FloatField)

    def test_integer_field(self):
        column = fields.IntegerField()
        self.assertIsInstance(column.type, sa.Integer)
        self.assertIsInstance(column.info["validators"][-2], django_validators.MinValueValidator)
        self.assertEqual(column.info["validators"][-2].limit_value, -2147483648)
        self.assertIsInstance(column.info["validators"][-1], django_validators.MaxValueValidator)
        self.assertEqual(column.info["validators"][-1].limit_value, 2147483647)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.IntegerField)

    def test_biginteger_field(self):
        column = fields.BigIntegerField()
        self.assertIsInstance(column.type, sa.BigInteger)
        self.assertIsInstance(column.info["validators"][-2], django_validators.MinValueValidator)
        self.assertEqual(column.info["validators"][-2].limit_value, -9223372036854775808)
        self.assertIsInstance(column.info["validators"][-1], django_validators.MaxValueValidator)
        self.assertEqual(column.info["validators"][-1].limit_value, 9223372036854775807)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.IntegerField)

    def test_smallinteger_field(self):
        column = fields.SmallIntegerField()
        self.assertIsInstance(column.type, sa.SmallInteger)
        self.assertIsInstance(column.info["validators"][-2], django_validators.MinValueValidator)
        self.assertEqual(column.info["validators"][-2].limit_value, -32768)
        self.assertIsInstance(column.info["validators"][-1], django_validators.MaxValueValidator)
        self.assertEqual(column.info["validators"][-1].limit_value, 32767)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.IntegerField)

    def test_nullboolean_field(self):
        column = fields.NullBooleanField()
        self.assertIsInstance(column.type, sa.Boolean)
        self.assertTrue(column.nullable)

        column = fields.NullBooleanField(constraint_name="test")
        self.assertEqual(column.type.name, "test")

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.NullBooleanField)

    def test_slug_field(self):
        column = fields.SlugField(label="test", length=100)
        self.assertIsInstance(column.type, sa.String)
        self.assertEqual(
            {
                "form_class": djangofields.SlugField,
                "label": "test",
                "required": False,
                "validators": [django_validators.validate_slug, mock.ANY],
            },
            column.info,
        )

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.SlugField)

    def test_text_field(self):
        column = fields.TextField()
        self.assertIsInstance(column.type, sa.Text)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.CharField)
        self.assertIsInstance(form_field.widget, djangoforms.Textarea)

    def test_time_field(self):
        column = fields.TimeField()
        self.assertIsInstance(column.type, sa.Time)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.TimeField)

    def test_timestamp_field(self):
        column = fields.TimestampField()
        self.assertIsInstance(column.type, sa.TIMESTAMP)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.DateTimeField)

    def test_url_field(self):
        column = fields.URLField(length=100)
        self.assertIsInstance(column.type, sa.String)
        self.assertIsInstance(column.info["validators"][0], django_validators.URLValidator)

        form_field = meta.column_info(column).formfield()
        self.assertIsInstance(form_field, djangofields.URLField)

    def test_binary_field(self):
        fields.BinaryField()

    def test_override_form_class(self):
        column = fields.EnumField(choices=["a", "b"], form_class=djangofields.ChoiceField)

        form_field = meta.column_info(column).formfield()
        self.assertIs(type(form_field), djangofields.ChoiceField)
        self.assertIsNot(type(form_field), djangofields.TypedChoiceField)

        column = fields.URLField(form_class=djangofields.CharField, length=100)
        form_field = meta.column_info(column).formfield()
        self.assertIs(type(form_field), djangofields.CharField)
        self.assertIsNot(type(form_field), djangofields.URLField)



"""Base model metadata things."""

import sqlalchemy as sa


class model_info_meta(type):
    """Model metadata singleton."""

    _registry = {}

    def __call__(cls, model, *args, **kwargs):
        obj = sa.inspect(model)

        if isinstance(obj, sa.orm.Mapper):
            model = obj.class_

        if isinstance(obj, sa.orm.state.InstanceState):
            model = obj.mapper.class_

        if model not in cls._registry:
            instance = super().__call__(model, *args, **kwargs)
            cls._registry[model] = instance

        return cls._registry[model]



from django.core.exceptions import ValidationError
from django_sorcery.db.transaction import TransactionContext

from ..base import TestCase
from ..testapp.models import Owner, db


class TestTransaction(TestCase):
    def setUp(self):
        super().setUp()
        Owner.query.delete()
        db.commit()
        db.remove()

    def tearDown(self):
        Owner.query.delete()
        db.commit()

    def test_transaction_context(self):
        transaction = TransactionContext(db)
        self.assertEqual(transaction.dbs, (db,))
        self.assertTrue(transaction.savepoint)
        self.assertIsNone(transaction.transactions)

    def test_decorator(self):
        @TransactionContext(db)
        def do_stuff():
            db.add(Owner(first_name="The", last_name="Dude"))

        do_stuff()
        self.assertEqual(Owner.objects.count(), 1)

    def test_decorator_invalid(self):
        @TransactionContext(db)
        def do_stuff():
            db.add(Owner(first_name="invalid", last_name="Dude"))

        with self.assertRaises(ValidationError):
            do_stuff()

        self.assertEqual(Owner.objects.count(), 0)

    def test_context_manager(self):
        with TransactionContext(db):
            db.add(Owner(first_name="The", last_name="Dude"))

        self.assertEqual(Owner.objects.count(), 1)

    def test_context_manager_invalid(self):
        with self.assertRaises(ValidationError):
            with TransactionContext(db):
                db.add(Owner(first_name="invalid", last_name="Dude"))

        self.assertEqual(Owner.objects.count(), 0)



from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django_sorcery import forms
from django_sorcery.views import edit

from ..base import TestCase
from ..testapp.models import Owner, db
from ..testapp.views import OwnerCreateViewWithForm


class TestCreateView(TestCase):
    def test_create(self):
        url = reverse("owner_create")

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.context_data["view"], edit.CreateView)
        self.assertIsInstance(response.context_data["form"], forms.ModelForm)

        response = self.client.post(url, {"first_name": "Randall", "last_name": "Munroe"})

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("owners_list"))

    def test_create_with_form_class(self):
        url = reverse("owner_create_form")

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.context_data["view"], edit.CreateView)
        self.assertTrue(type(response.context_data["form"]) is OwnerCreateViewWithForm.form_class)

    def test_create_bad_field_form_config(self):
        url = reverse("owner_create_field_form")

        with self.assertRaises(ImproperlyConfigured) as ctx:
            self.client.get(url)

        self.assertEqual(str(ctx.exception), "Specifying both 'fields' and 'form_class' is not permitted.")

    def test_create_get_success_url_bad_config(self):
        view = OwnerCreateViewWithForm()

        with self.assertRaises(ImproperlyConfigured) as ctx:
            view.get_success_url()

        self.assertEqual(
            str(ctx.exception), "No URL to redirect to. Either provide a url or override this function to return a url"
        )

    def test_create_form_invalid(self):
        url = reverse("vehicle_create")

        response = self.client.post(url, {})

        self.assertEqual(response.status_code, 200)

        form = response.context_data["form"]
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {"type": ["This field is required."], "is_used": ["This field is required."]})
        self.assertHTMLEqual(response.content.decode(), form.as_p())


class TestUpdateView(TestCase):
    def test_update(self):
        owner = Owner(first_name="Radnall", last_name="Munroe")
        db.add(owner)
        db.flush()

        url = reverse("owner_update", kwargs={"id": owner.id})

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.context_data["view"], edit.UpdateView)
        self.assertIsInstance(response.context_data["form"], forms.ModelForm)
        form = response.context_data["form"]
        view = response.context_data["view"]
        self.assertHTMLEqual(response.content.decode(), form.as_p())
        self.assertEqual(view.object, owner)

        response = self.client.post(url, {"first_name": "Patrick", "last_name": "Bateman"})

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("owners_list"))

        self.assertEqual(owner.first_name, "Patrick")
        self.assertEqual(owner.last_name, "Bateman")


class TestDeleteView(TestCase):
    def test_delete(self):
        owner = Owner(first_name="Radnall", last_name="Munroe")
        db.add(owner)
        db.flush()

        url = reverse("owner_delete", kwargs={"id": owner.id})

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.context_data["view"], edit.DeleteView)
        self.assertEqual(response.context_data["object"], owner)

        response = self.client.post(url)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("owners_list"))

        state = db.inspect(owner)
        self.assertTrue(state.deleted)

    def test_delete_get_success_url(self):
        view = edit.DeletionMixin()

        with self.assertRaises(ImproperlyConfigured):
            view.get_success_url()



from django.core.exceptions import ImproperlyConfigured
from django.http import Http404
from django_sorcery import shortcuts

from .base import TestCase
from .testapp.models import Owner, db


class TestShortcuts(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all(
            [
                Owner(first_name="Test 1", last_name="Owner 1"),
                Owner(first_name="Test 2", last_name="Owner 2"),
                Owner(first_name="Test 3", last_name="Owner 3"),
                Owner(first_name="Test 4", last_name="Owner 4"),
            ]
        )
        db.flush()

    def test_get_object_or_404(self):
        owner = Owner.objects.first()

        obj = shortcuts.get_object_or_404(Owner, id=owner.id)
        self.assertIsNotNone(obj)

        with self.assertRaises(Http404):
            shortcuts.get_object_or_404(Owner, id=0)

        obj = shortcuts.get_object_or_404(Owner.query, id=owner.id)
        self.assertIsNotNone(obj)

        with self.assertRaises(Http404):
            shortcuts.get_object_or_404(Owner.query, id=0)

        with self.assertRaises(ImproperlyConfigured):

            class Dummy:
                pass

            shortcuts.get_object_or_404(Dummy, id=0)

    def test_get_list_or_404(self):
        obj = shortcuts.get_list_or_404(Owner)
        self.assertTrue(list(obj))

        with self.assertRaises(Http404):
            shortcuts.get_list_or_404(Owner, Owner.first_name.ilike("XXX"))

        obj = shortcuts.get_list_or_404(Owner, first_name="Test 1")
        self.assertTrue(list(obj))

        with self.assertRaises(Http404):
            shortcuts.get_list_or_404(Owner, first_name="XXX")



"""Common mixins used in models."""

from . import meta


class CleanMixin:
    """Mixin for adding django-style ``full_clean`` validation to any object.

    Base model in :py:class:`..sqlalchemy.SQLAlchemy` already uses this mixin applied.

    For example::

        class Address(db.Model):
            city = db.Column(db.String(20))
            state = db.Column(db.String(2))
            date = db.Column(db.Date())

            validators = [
                ValidateTogetherModelFields(["city", "state"]),
            ]

            def clean_date(self):
                if self.date > datetime.date.today():
                    raise ValidationError("Cant pick future date")

            def clean(self):
                if self.date.year < 1776 and self.state == "NY":
                    raise ValidationError("NY state did not exist before 1776")
    """

    def clean(self, **kwargs):
        """Hook for adding custom model validations before model is flushed.

        Should raise ``ValidationError`` if any errors are found.
        """

    def clean_fields(self, exclude=None, **kwargs):
        """Clean all fields on object."""
        meta.model_info(self).clean_fields(self, exclude=exclude, **kwargs)

    def _get_properties_for_validation(self):
        """Needs to be implemented to return all properties for the object."""

    def clean_nested_fields(self, exclude=None, **kwargs):
        """Clean all nested fields which includes composites."""
        meta.model_info(self).clean_nested_fields(self, exclude=exclude, **kwargs)

    def _get_nested_objects_for_validation(self):
        """Needs to be implemented to return all nested objects."""

    def clean_relation_fields(self, exclude, **kwargs):
        """Clean all relation fields."""
        meta.model_info(self).clean_relation_fields(self, exclude=exclude, **kwargs)

    def _get_relation_objects_for_validation(self):
        """Needs to be implemented to return all relation objects."""

    def run_validators(self, **kwargs):
        """Check all model validators registered on ``validators``
        attribute."""
        meta.model_info(self).run_validators(self, **kwargs)

    def full_clean(self, exclude=None, **kwargs):
        """Run model's full clean chain.

        This will run all of these in this order:

        * will validate all columns by using ``clean_<column>`` methods
        * will validate all nested objects (e.g. composites) with ``full_clean``
        * will run through all registered validators on ``validators`` attribute
        * will run full model validation with ``self.clean()``
        * if ``recursive`` kwarg is provided, will recursively clean all relations.
          Useful when all models need to be explicitly cleaned without flushing to DB.
        """
        meta.model_info(self).full_clean(self, exclude=exclude, **kwargs)



"""pytest plugins."""

import pytest

from .db.profiler import SQLAlchemyProfiler
from .testing import Transact


@pytest.fixture(scope="function")
def sqlalchemy_profiler():
    """pytest fixture for sqlalchemy profiler."""
    with SQLAlchemyProfiler() as profiler:  # pragma: nocover
        yield profiler


@pytest.fixture(scope="function")
def transact():
    """pytest transact fixture for sqlalchemy."""
    with Transact() as t:
        yield t



from django.core.exceptions import ImproperlyConfigured
from django.http import Http404
from django.urls import reverse
from django_sorcery import forms, viewsets

from ..base import TestCase
from ..testapp.models import Owner, db


class TestListModelMixin(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all(
            [
                Owner(id=1, first_name="Test 1", last_name="Owner 1"),
                Owner(id=2, first_name="Test 2", last_name="Owner 2"),
                Owner(id=3, first_name="Test 3", last_name="Owner 3"),
                Owner(id=4, first_name="Test 4", last_name="Owner 4"),
            ]
        )
        db.flush()

    def test_list(self):
        class OwnerViewSet(viewsets.ListModelMixin, viewsets.GenericViewSet):
            model = Owner

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.get("/")
        viewset.action = "list"

        response = viewset.list(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.context_data,
            {
                "is_paginated": False,
                "object_list": viewset.object_list,
                "owner_list": viewset.object_list,
                "page_obj": None,
                "paginator": None,
                "view": viewset,
            },
        )

    def test_list_allow_no_empty(self):
        class OwnerViewSet(viewsets.ListModelMixin, viewsets.GenericViewSet):
            model = Owner
            allow_empty = False
            paginate_by = 10

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.get("/")
        viewset.action = "list"

        response = viewset.list(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.context_data,
            {
                "is_paginated": False,
                "object_list": viewset.object_list.all(),
                "owner_list": viewset.object_list.all(),
                "page_obj": response.context_data["page_obj"],
                "paginator": response.context_data["paginator"],
                "view": viewset,
            },
        )

    def test_list_allow_no_empty_with_no_owners(self):
        db.rollback()

        class OwnerViewSet(viewsets.ListModelMixin, viewsets.GenericViewSet):
            model = Owner
            allow_empty = False
            paginate_by = 10

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.get("/")
        viewset.action = "list"

        with self.assertRaises(Http404):
            viewset.list(viewset.request)


class TestRetieveModelMixin(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all(
            [
                Owner(id=1, first_name="Test 1", last_name="Owner 1"),
                Owner(id=2, first_name="Test 2", last_name="Owner 2"),
                Owner(id=3, first_name="Test 3", last_name="Owner 3"),
                Owner(id=4, first_name="Test 4", last_name="Owner 4"),
            ]
        )
        db.flush()

    def test_retrieve(self):
        class OwnerViewSet(viewsets.RetrieveModelMixin, viewsets.GenericViewSet):
            model = Owner

        viewset = OwnerViewSet()
        viewset.kwargs = {"id": 1}
        viewset.request = self.factory.get("/")
        viewset.action = "retrieve"

        response = viewset.retrieve(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.context_data, {"object": viewset.object, "owner": viewset.object, "view": viewset}
        )

        kwargs = viewset.get_url_kwargs(viewset.object)
        self.assertDictEqual(kwargs, {"id": 1})


class TestCreateModelMixin(TestCase):
    def test_new_require_fields_or_form_class(self):
        class OwnerViewSet(viewsets.CreateModelMixin, viewsets.GenericViewSet):
            model = Owner
            fields = "__all__"
            form_class = forms.modelform_factory(Owner, fields="__all__", session=db)

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.get("/")
        viewset.action = "new"

        with self.assertRaises(ImproperlyConfigured):
            viewset.new(viewset.request)

    def test_new_with_form_class(self):
        form_cls = forms.modelform_factory(Owner, fields="__all__", session=db)

        class OwnerViewSet(viewsets.CreateModelMixin, viewsets.GenericViewSet):
            model = Owner
            form_class = form_cls

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.get("/")
        viewset.action = "new"

        response = viewset.new(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.context_data, {"form": response.context_data["form"], "view": viewset})
        self.assertIsInstance(response.context_data["form"], form_cls)

    def test_new(self):
        class OwnerViewSet(viewsets.CreateModelMixin, viewsets.GenericViewSet):
            model = Owner
            fields = "__all__"

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.get("/")
        viewset.action = "new"

        response = viewset.new(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.context_data, {"form": response.context_data["form"], "view": viewset})

    def test_create_bad_config(self):
        class OwnerViewSet(viewsets.CreateModelMixin, viewsets.GenericViewSet):
            model = Owner
            fields = (Owner.first_name.key, Owner.last_name.key)

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.post("/", data={"first_name": "first_name", "last_name": "last_name"})
        viewset.action = "create"

        with self.assertRaises(ImproperlyConfigured):
            viewset.create(viewset.request)

    def test_create(self):
        class OwnerViewSet(viewsets.CreateModelMixin, viewsets.GenericViewSet):
            model = Owner
            fields = (Owner.first_name.key, Owner.last_name.key)
            success_url = "/viewsets/owners/{id}/"

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.post("/", data={"first_name": "first_name", "last_name": "last_name"})
        viewset.action = "create"

        response = viewset.create(viewset.request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.get("location"), reverse("owner-detail", args=[viewset.object.id]))
        self.assertEqual(viewset.object.first_name, "first_name")
        self.assertEqual(viewset.object.last_name, "last_name")

    def test_create_invalid(self):
        form_cls = forms.modelform_factory(Owner, fields=("first_name", "last_name"), session=db)
        form_cls.base_fields["first_name"].required = True
        form_cls.base_fields["last_name"].required = True

        class OwnerViewSet(viewsets.CreateModelMixin, viewsets.GenericViewSet):
            model = Owner
            form_class = form_cls
            success_url = "/viewsets/owners/{id}/"

        viewset = OwnerViewSet()
        viewset.kwargs = {}
        viewset.request = self.factory.post("/", data={})
        viewset.action = "create"

        response = viewset.create(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.context_data["form"].errors,
            {"first_name": ["This field is required."], "last_name": ["This field is required."]},
        )


class TestUpdateModelMixin(TestCase):
    def setUp(self):
        super().setUp()
        db.add(Owner(id=1, first_name="Test 1", last_name="Owner 1"))
        db.flush()

    def test_edit(self):
        class OwnerViewSet(viewsets.UpdateModelMixin, viewsets.GenericViewSet):
            model = Owner
            fields = ("first_name", "last_name")

        viewset = OwnerViewSet()
        viewset.kwargs = {"id": 1}
        viewset.request = self.factory.get("/")
        viewset.action = "edit"

        response = viewset.edit(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.context_data,
            {"form": response.context_data["form"], "owner": viewset.object, "object": viewset.object, "view": viewset},
        )

    def test_update(self):
        class OwnerViewSet(viewsets.UpdateModelMixin, viewsets.GenericViewSet):
            model = Owner
            fields = ("first_name", "last_name")
            success_url = "/viewsets/owners/{id}/"

        viewset = OwnerViewSet()
        viewset.kwargs = {"id": 1}
        viewset.request = self.factory.post("/", data={"first_name": "edit first", "last_name": "edit last"})
        viewset.action = "update"

        response = viewset.update(viewset.request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.get("location"), "/viewsets/owners/1/")


class TestDeleteModelMixin(TestCase):
    def setUp(self):
        super().setUp()
        db.add(Owner(id=1, first_name="Test 1", last_name="Owner 1"))
        db.flush()

    def test_confirm_destroy(self):
        class OwnerViewSet(viewsets.DeleteModelMixin, viewsets.GenericViewSet):
            model = Owner

        viewset = OwnerViewSet()
        viewset.kwargs = {"id": 1}
        viewset.request = self.factory.get("/")
        viewset.action = "confirm_destory"

        response = viewset.confirm_destroy(viewset.request)

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.context_data, {"owner": viewset.object, "object": viewset.object, "view": viewset}
        )

    def test_destroy(self):
        class OwnerViewSet(viewsets.DeleteModelMixin, viewsets.GenericViewSet):
            model = Owner
            destroy_success_url = "/viewsets/owners/"

        viewset = OwnerViewSet()
        viewset.kwargs = {"id": 1}
        viewset.request = self.factory.post("/")
        viewset.action = "destory"

        response = viewset.destroy(viewset.request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.get("location"), "/viewsets/owners/")
        self.assertIsNone(Owner.query.get(1))

    def test_destroy_no_success_url(self):
        class OwnerViewSet(viewsets.DeleteModelMixin, viewsets.GenericViewSet):
            model = Owner

        viewset = OwnerViewSet()
        viewset.kwargs = {"id": 1}
        viewset.request = self.factory.post("/")
        viewset.action = "destory"

        with self.assertRaises(ImproperlyConfigured):
            viewset.destroy(viewset.request)
